from __future__ import annotations

from enum import Enum
from typing import Dict, Optional, Tuple, Union

from boa3.constants import FOUR_BYTES_MAX_VALUE
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String


class Opcode(bytes, Enum):
    # region Constants

    # Pushes a 1-byte signed integer onto the stack.
    PUSHINT8 = b'\x00'
    # Pushes a 2-byte signed integer onto the stack.
    PUSHINT16 = b'\x01'
    # Pushes a 4-byte signed integer onto the stack.
    PUSHINT32 = b'\x02'
    # Pushes a 8-byte signed integer onto the stack.
    PUSHINT64 = b'\x03'
    # Pushes a 16-byte signed integer onto the stack.
    PUSHINT128 = b'\x04'
    # Pushes a 32-byte signed integer onto the stack.
    PUSHINT256 = b'\x05'

    # Converts the 4-bytes offset to a Pointer, and pushes it onto the stack.
    PUSHA = b'\x0A'
    # The item null is pushed onto the stack.
    PUSHNULL = b'\x0B'
    # The next byte contains the number of bytes to be pushed onto the stack.
    PUSHDATA1 = b'\x0C'
    # The next two bytes contain the number of bytes to be pushed onto the stack.
    PUSHDATA2 = b'\x0D'
    # The next four bytes contain the number of bytes to be pushed onto the stack.
    PUSHDATA4 = b'\x0E'

    @staticmethod
    def get_pushdata_and_data(bytestring: Union[str, bytes]) -> Tuple[Opcode, bytes]:
        """
        Gets the push opcode and data to the respective str ot bytes value

        :param bytestring: value that will be pushed
        :return: the respective opcode and its required data
        :rtype: Tuple[Opcode, bytes]
        """
        if isinstance(bytestring, str):
            bytestring = String(bytestring).to_bytes()

        bytes_size = Integer(len(bytestring)).to_byte_array(min_length=1)
        if len(bytes_size) == 1:
            opcode = Opcode.PUSHDATA1
        elif len(bytes_size) == 2:
            opcode = Opcode.PUSHDATA2
        else:
            if len(bytes_size) > 4:
                bytestring = bytestring[:FOUR_BYTES_MAX_VALUE]
            bytes_size = Integer(len(bytestring)).to_byte_array(min_length=4)
            opcode = Opcode.PUSHDATA4

        data = bytes_size + bytestring
        return opcode, data

    @staticmethod
    def get_pushdata_and_data_from_size(data_size: int) -> Tuple[Opcode, bytes]:
        bytes_size = Integer(data_size).to_byte_array(min_length=1)
        if len(bytes_size) == 1:
            opcode = Opcode.PUSHDATA1
        elif len(bytes_size) == 2:
            opcode = Opcode.PUSHDATA2
        else:
            bytes_size = FOUR_BYTES_MAX_VALUE
            opcode = Opcode.PUSHDATA4

        return opcode, bytes_size

    @staticmethod
    def get_literal_push(integer: int) -> Optional[Opcode]:
        """
        Gets the push opcode to the respective integer

        :param integer: value that will be pushed
        :return: the respective opcode
        :rtype: Opcode or None
        """
        if -1 <= integer <= 16:
            opcode_value: int = Integer.from_bytes(Opcode.PUSH0) + integer
            return Opcode(Integer(opcode_value).to_byte_array())
        else:
            return None

    @staticmethod
    def get_push_and_data(integer: int) -> Tuple[Opcode, bytes]:
        """
        Gets the push opcode and data to the respective integer

        :param integer: value that will be pushed
        :return: the respective opcode and its required data
        :rtype: Tuple[Opcode, bytes]
        """
        opcode = Opcode.get_literal_push(integer)
        if isinstance(opcode, Opcode):
            return opcode, b''
        else:
            data = Integer(integer).to_byte_array(signed=True, min_length=1)
            if len(data) == 1:
                opcode = Opcode.PUSHINT8
            elif len(data) == 2:
                opcode = Opcode.PUSHINT16
            elif len(data) <= 4:
                data = Integer(integer).to_byte_array(signed=True, min_length=4)
                opcode = Opcode.PUSHINT32
            elif len(data) <= 8:
                data = Integer(integer).to_byte_array(signed=True, min_length=8)
                opcode = Opcode.PUSHINT64
            elif len(data) <= 16:
                data = Integer(integer).to_byte_array(signed=True, min_length=16)
                opcode = Opcode.PUSHINT128
            else:
                data = data[:32]
                opcode = Opcode.PUSHINT256

            return opcode, data

    # The number -1 is pushed onto the stack.
    PUSHM1 = b'\x0F'
    # The number 0 is pushed onto the stack.
    PUSH0 = b'\x10'
    # The number 1 is pushed onto the stack.
    PUSH1 = b'\x11'
    # The number 2 is pushed onto the stack.
    PUSH2 = b'\x12'
    # The number 3 is pushed onto the stack.
    PUSH3 = b'\x13'
    # The number 4 is pushed onto the stack.
    PUSH4 = b'\x14'
    # The number 5 is pushed onto the stack.
    PUSH5 = b'\x15'
    # The number 6 is pushed onto the stack.
    PUSH6 = b'\x16'
    # The number 7 is pushed onto the stack.
    PUSH7 = b'\x17'
    # The number 8 is pushed onto the stack.
    PUSH8 = b'\x18'
    # The number 9 is pushed onto the stack.
    PUSH9 = b'\x19'
    # The number 10 is pushed onto the stack.
    PUSH10 = b'\x1A'
    # The number 11 is pushed onto the stack.
    PUSH11 = b'\x1B'
    # The number 12 is pushed onto the stack.
    PUSH12 = b'\x1C'
    # The number 13 is pushed onto the stack.
    PUSH13 = b'\x1D'
    # The number 14 is pushed onto the stack.
    PUSH14 = b'\x1E'
    # The number 15 is pushed onto the stack.
    PUSH15 = b'\x1F'
    # The number 16 is pushed onto the stack.
    PUSH16 = b'\x20'

    # endregion

    # region Flow control

    def has_larger_opcode(self) -> bool:
        return self in self.__larger_opcode

    def get_larger_opcode(self):
        """
        Gets the large opcode to the standard opcode

        :return: the respective opcode
        :rtype: Opcode or None
        """
        opcode_map = self.__larger_opcode
        if self in opcode_map:
            return opcode_map[self]
        elif self in opcode_map.values():
            return self
        else:
            return None

    @property
    def __larger_opcode(self) -> Dict[Opcode, Opcode]:
        """
        A map of each opcode with its larger equivalent

        :return: a dictionary that maps each opcode to its larger equivalent.
        """
        opcodes = {
            self.JMP: self.JMP_L,
            self.JMPIF: self.JMPIF_L,
            self.JMPIFNOT: self.JMPIFNOT_L,
            self.JMPEQ: self.JMPEQ_L,
            self.JMPNE: self.JMPNE_L,
            self.JMPGT: self.JMPGT_L,
            self.JMPGE: self.JMPGE_L,
            self.JMPLT: self.JMPLT_L,
            self.JMPLE: self.JMPLE_L,
            self.CALL: self.CALL_L,
            self.TRY: self.TRY_L,
            self.ENDTRY: self.ENDTRY_L
        }
        return opcodes

    # The NOP operation does nothing. It is intended to fill in space if opcodes are patched.
    NOP = b'\x21'
    # Unconditionally transfers control to a target instruction. The target instruction is represented as a 1-byte
    # signed offset from the beginning of the current instruction.
    JMP = b'\x22'
    # Unconditionally transfers control to a target instruction. The target instruction is represented as a 4-bytes
    # signed offset from the beginning of the current instruction.
    JMP_L = b'\x23'
    # Transfers control to a target instruction if the value is True, not null, or non-zero. The target instruction
    # is represented as a 1-byte signed offset from the beginning of the current instruction.
    JMPIF = b'\x24'
    # Transfers control to a target instruction if the value is True, not null, or non-zero. The target instruction
    # is represented as a 4-bytes signed offset from the beginning of the current instruction.
    JMPIF_L = b'\x25'
    # Transfers control to a target instruction if the value is False, a null reference,
    # or zero. The target instruction is represented as a 1-byte signed offset from the beginning of the current
    # instruction.
    JMPIFNOT = b'\x26'
    # Transfers control to a target instruction if the value is False, a null reference,
    # or zero. The target instruction is represented as a 4-bytes signed offset from the beginning of the current
    # instruction.
    JMPIFNOT_L = b'\x27'
    # Transfers control to a target instruction if two values are equal. The target instruction is represented as a
    # 1-byte signed offset from the beginning of the current instruction.
    JMPEQ = b'\x28'
    # Transfers control to a target instruction if two values are equal. The target instruction is represented as a
    # 4-bytes signed offset from the beginning of the current instruction.
    JMPEQ_L = b'\x29'
    # Transfers control to a target instruction when two values are not equal. The target instruction is represented
    # as a 1-byte signed offset from the beginning of the current instruction.
    JMPNE = b'\x2A'
    # Transfers control to a target instruction when two values are not equal. The target instruction is represented
    # as a 4-bytes signed offset from the beginning of the current instruction.
    JMPNE_L = b'\x2B'
    # Transfers control to a target instruction if the first value is greater than the second value. The target
    # instruction is represented as a 1-byte signed offset from the beginning of the current instruction.
    JMPGT = b'\x2C'
    # Transfers control to a target instruction if the first value is greater than the second value. The target
    # instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.
    JMPGT_L = b'\x2D'
    # Transfers control to a target instruction if the first value is greater than or equal to the second value. The
    # target instruction is represented as a 1-byte signed offset from the beginning of the current instruction.
    JMPGE = b'\x2E'
    # Transfers control to a target instruction if the first value is greater than or equal to the second value. The
    # target instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.
    JMPGE_L = b'\x2F'
    # Transfers control to a target instruction if the first value is less than the second value. The target
    # instruction is represented as a 1-byte signed offset from the beginning of the current instruction.
    JMPLT = b'\x30'
    # Transfers control to a target instruction if the first value is less than the second value. The target
    # instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.
    JMPLT_L = b'\x31'
    # Transfers control to a target instruction if the first value is less than or equal to the second value. The
    # target instruction is represented as a 1-byte signed offset from the beginning of the current instruction.
    JMPLE = b'\x32'
    # Transfers control to a target instruction if the first value is less than or equal to the second value. The
    # target instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.
    JMPLE_L = b'\x33'

    @property
    def is_jump(self) -> bool:
        return self.JMP <= self <= self.JMPLE_L

    @staticmethod
    def get_jump_and_data(opcode: Opcode, integer: int, jump_through: bool = False) -> Tuple[Opcode, bytes]:
        """
        Gets the jump opcode and data to the respective integer

        :param opcode: which jump will be used
        :param integer: number of bytes that'll be jumped
        :param jump_through: whether it should go over the instructions or not
        :return: the respective opcode and its required data
        :rtype: Tuple[Opcode, bytes]
        """
        if not opcode.is_jump:
            opcode = Opcode.JMP

        from boa3.neo.vm.opcode.OpcodeInfo import OpcodeInfo
        opcode_info = OpcodeInfo.get_info(opcode)
        arg_size = opcode_info.data_len
        jmp_arg = Integer(arg_size + integer + jump_through).to_byte_array(min_length=arg_size) if integer > 0 \
            else Integer(integer).to_byte_array(min_length=arg_size)

        if len(jmp_arg) > opcode_info.max_data_len and opcode.has_larger_opcode():
            opcode = opcode.get_larger_opcode()
            opcode_info = OpcodeInfo.get_info(opcode)
            arg_size = opcode_info.data_len
            jmp_arg = Integer(arg_size + integer + jump_through).to_byte_array(min_length=arg_size) if integer > 0 \
                else Integer(integer).to_byte_array(min_length=arg_size)
            jmp_opcode = opcode
        else:
            jmp_opcode = opcode

        jmp_arg = jmp_arg[-arg_size:]

        return jmp_opcode, jmp_arg

    # Calls the function at the target address which is represented as a 1-byte signed offset from the beginning of
    # the current instruction.
    CALL = b'\x34'
    # Calls the function at the target address which is represented as a 4-bytes signed offset from the beginning of
    # the current instruction.
    CALL_L = b'\x35'
    # Pop the address of a function from the stack, and call the function.
    CALLA = b'\x36'
    # Calls the function which is described by the token.
    CALLT = b'\x37'
    # It turns the vm state to FAULT immediately, and cannot be caught.
    ABORT = b'\x38'
    # Pop the top value of the stack, if it false, then exit vm execution and set vm state to FAULT.
    ASSERT = b'\x39'
    # Pop the top value of the stack, and throw it.
    THROW = b'\x3A'
    # TRY CatchOffset(sbyte) FinallyOffset(sbyte). If there's no catch body, set CatchOffset 0. If there's no finally
    # body, set FinallyOffset 0.
    TRY = b'\x3B'
    # TRY_L CatchOffset(int) FinallyOffset(int). If there's no catch body, set CatchOffset 0. If there's no finally
    # body, set FinallyOffset 0.
    TRY_L = b'\x3C'
    # Ensures that the appropriate surrounding finally blocks are executed. And then unconditionally transfers
    # control to the specific target instruction, represented as a 1-byte signed offset from the beginning of the
    # current instruction.
    ENDTRY = b'\x3D'
    # Ensures that the appropriate surrounding finally blocks are executed. And then unconditionally transfers
    # control to the specific target instruction, represented as a 4-byte signed offset from the beginning of the
    # current instruction.
    ENDTRY_L = b'\x3E'
    # End finally, If no exception happen or be catched, vm will jump to the target instruction of ENDTRY/ENDTRY_L.
    # Otherwise vm will rethrow the exception to upper layer.
    ENDFINALLY = b'\x3F'
    # Returns from the current method.
    RET = b'\x40'
    # Calls to an interop service.
    SYSCALL = b'\x41'

    def has_target(self) -> bool:
        return self.JMP <= self <= self.CALL_L or self.TRY <= self < self.ENDFINALLY

    # endregion

    # region Stack

    # Puts the number of stack items onto the stack.
    DEPTH = b'\x43'
    # Removes the top stack item.
    DROP = b'\x45'
    # Removes the second-to-top stack item.
    NIP = b'\x46'
    # The item n back in the main stack is removed.
    XDROP = b'\x48'
    # Clear the stack
    CLEAR = b'\x49'

    @staticmethod
    def get_drop(position: int) -> Optional[Opcode]:
        """
        Gets the opcode to remove the item n back in the stack

        :param position: index of the variable
        :return: the respective opcode
        :rtype: Opcode
        """
        duplicate_item = {
            1: Opcode.DROP,
            2: Opcode.NIP
        }

        if position > 0:
            if position in duplicate_item:
                return duplicate_item[position]
            else:
                return Opcode.XDROP

    # Duplicates the top stack item.
    DUP = b'\x4A'
    # Copies the second-to-top stack item to the top.
    OVER = b'\x4B'
    # The item n back in the stack is copied to the top.
    PICK = b'\x4D'

    @staticmethod
    def get_dup(position: int) -> Optional[Opcode]:
        """
        Gets the opcode to duplicate the item n back in the stack

        :param position: index of the variable
        :return: the respective opcode
        :rtype: Opcode
        """
        duplicate_item = {
            1: Opcode.DUP,
            2: Opcode.OVER
        }

        if position >= 0:
            if position in duplicate_item:
                return duplicate_item[position]
            else:
                return Opcode.PICK

    # The item at the top of the stack is copied and inserted before the second-to-top item.
    TUCK = b'\x4E'
    # The top two items on the stack are swapped.
    SWAP = b'\x50'
    # The top three items on the stack are rotated to the left.
    ROT = b'\x51'
    # The item n back in the stack is moved to the top.
    ROLL = b'\x52'
    # Reverse the order of the top 3 items on the stack.
    REVERSE3 = b'\x53'
    # Reverse the order of the top 4 items on the stack.
    REVERSE4 = b'\x54'
    # Pop the number N on the stack, and reverse the order of the top N items on the stack.
    REVERSEN = b'\x55'

    @staticmethod
    def get_reverse(no_stack_items: int) -> Optional[Opcode]:
        """
        Gets the opcode to reverse n items on the stack

        :param no_stack_items: index of the variable
        :return: the respective opcode
        :rtype: Opcode
        """
        reverse_stack = {
            2: Opcode.SWAP,
            3: Opcode.REVERSE3,
            4: Opcode.REVERSE4
        }

        if no_stack_items >= 0:
            if no_stack_items in reverse_stack:
                return reverse_stack[no_stack_items]
            else:
                return Opcode.REVERSEN

    # endregion

    # region Slot

    # Initialize the static field list for the current execution context.
    INITSSLOT = b'\x56'
    # Initialize the argument slot and the local variable list for the current execution context.
    INITSLOT = b'\x57'

    @staticmethod
    def get_store(index: int, local: bool, is_arg: bool = False) -> Opcode:
        """
        Gets the opcode to store the variable

        :param index: index of the variable
        :param local: identifies if the variable is local or global
        :param is_arg: identifies if the variable is an argument of a function. False if local is False.
        :return: the respective opcode
        :rtype: Opcode
        """
        if not local:
            is_arg = False

        if 0 <= index <= 6:
            if is_arg:
                opcode_value: int = Integer.from_bytes(Opcode.STARG0) + index
            elif local:
                opcode_value: int = Integer.from_bytes(Opcode.STLOC0) + index
            else:
                opcode_value: int = Integer.from_bytes(Opcode.STSFLD0) + index
            return Opcode(Integer(opcode_value).to_byte_array())
        else:
            if is_arg:
                return Opcode.STARG
            elif local:
                return Opcode.STLOC
            else:
                return Opcode.STSFLD

    @staticmethod
    def get_load(index: int, local: bool, is_arg: bool = False) -> Opcode:
        """
        Gets the opcode to load the variable

        :param index: index of the variable
        :param local: identifies if the variable is local or global
        :param is_arg: identifies if the variable is an argument of a function. False if local is False.
        :return: the respective opcode
        :rtype: Opcode
        """
        if not local:
            is_arg = False

        if 0 <= index <= 6:
            if is_arg:
                opcode_value: int = Integer.from_bytes(Opcode.LDARG0) + index
            elif local:
                opcode_value: int = Integer.from_bytes(Opcode.LDLOC0) + index
            else:
                opcode_value: int = Integer.from_bytes(Opcode.LDSFLD0) + index
            return Opcode(Integer(opcode_value).to_byte_array())
        else:
            if is_arg:
                return Opcode.LDARG
            elif local:
                return Opcode.LDLOC
            else:
                return Opcode.LDSFLD

    @property
    def is_load_slot(self) -> bool:
        return (self.LDSFLD0 <= self <= self.LDSFLD
                or self.LDLOC0 <= self <= self.LDLOC
                or self.LDARG0 <= self <= self.LDARG)

    @staticmethod
    def get_store_from_load(load_opcode) -> Optional[Opcode]:
        """
        Gets the store slot opcode equivalent to the given load slot opcode.

        :param load_opcode: load opcode
        :type load_opcode: Opcode
        :return: equivalent store opcode if the given opcode is a load slot. Otherwise, returns None
        :rtype: Opcode or None
        """
        if load_opcode.is_load_slot:
            opcode_value: int = Integer.from_bytes(load_opcode) + 8
            return Opcode(Integer(opcode_value).to_byte_array())

    # Loads the static field at index 0 onto the evaluation stack.
    LDSFLD0 = b'\x58'
    # Loads the static field at index 1 onto the evaluation stack.
    LDSFLD1 = b'\x59'
    # Loads the static field at index 2 onto the evaluation stack.
    LDSFLD2 = b'\x5A'
    # Loads the static field at index 3 onto the evaluation stack.
    LDSFLD3 = b'\x5B'
    # Loads the static field at index 4 onto the evaluation stack.
    LDSFLD4 = b'\x5C'
    # Loads the static field at index 5 onto the evaluation stack.
    LDSFLD5 = b'\x5D'
    # Loads the static field at index 6 onto the evaluation stack.
    LDSFLD6 = b'\x5E'
    # Loads the static field at a specified index onto the evaluation stack. The index is represented as a 1-byte
    # unsigned integer.
    LDSFLD = b'\x5F'
    # Stores the value on top of the evaluation stack in the static field list at index 0.
    STSFLD0 = b'\x60'
    # Stores the value on top of the evaluation stack in the static field list at index 1.
    STSFLD1 = b'\x61'
    # Stores the value on top of the evaluation stack in the static field list at index 2.
    STSFLD2 = b'\x62'
    # Stores the value on top of the evaluation stack in the static field list at index 3.
    STSFLD3 = b'\x63'
    # Stores the value on top of the evaluation stack in the static field list at index 4.
    STSFLD4 = b'\x64'
    # Stores the value on top of the evaluation stack in the static field list at index 5.
    STSFLD5 = b'\x65'
    # Stores the value on top of the evaluation stack in the static field list at index 6.
    STSFLD6 = b'\x66'
    # Stores the value on top of the evaluation stack in the static field list at a specified index. The index is
    # represented as a 1-byte unsigned integer.
    STSFLD = b'\x67'
    # Loads the local variable at index 0 onto the evaluation stack.
    LDLOC0 = b'\x68'
    # Loads the local variable at index 1 onto the evaluation stack.
    LDLOC1 = b'\x69'
    # Loads the local variable at index 2 onto the evaluation stack.
    LDLOC2 = b'\x6A'
    # Loads the local variable at index 3 onto the evaluation stack.
    LDLOC3 = b'\x6B'
    # Loads the local variable at index 4 onto the evaluation stack.
    LDLOC4 = b'\x6C'
    # Loads the local variable at index 5 onto the evaluation stack.
    LDLOC5 = b'\x6D'
    # Loads the local variable at index 6 onto the evaluation stack.
    LDLOC6 = b'\x6E'
    # Loads the local variable at a specified index onto the evaluation stack. The index is represented as a 1-byte
    # unsigned integer.
    LDLOC = b'\x6F'
    # Stores the value on top of the evaluation stack in the local variable list at index 0.
    STLOC0 = b'\x70'
    # Stores the value on top of the evaluation stack in the local variable list at index 1.
    STLOC1 = b'\x71'
    # Stores the value on top of the evaluation stack in the local variable list at index 2.
    STLOC2 = b'\x72'
    # Stores the value on top of the evaluation stack in the local variable list at index 3.
    STLOC3 = b'\x73'
    # Stores the value on top of the evaluation stack in the local variable list at index 4.
    STLOC4 = b'\x74'
    # Stores the value on top of the evaluation stack in the local variable list at index 5.
    STLOC5 = b'\x75'
    # Stores the value on top of the evaluation stack in the local variable list at index 6.
    STLOC6 = b'\x76'
    # Stores the value on top of the evaluation stack in the local variable list at a specified index. The index is
    # represented as a 1-byte unsigned integer.
    STLOC = b'\x77'
    # Loads the argument at index 0 onto the evaluation stack.
    LDARG0 = b'\x78'
    # Loads the argument at index 1 onto the evaluation stack.
    LDARG1 = b'\x79'
    # Loads the argument at index 2 onto the evaluation stack.
    LDARG2 = b'\x7A'
    # Loads the argument at index 3 onto the evaluation stack.
    LDARG3 = b'\x7B'
    # Loads the argument at index 4 onto the evaluation stack.
    LDARG4 = b'\x7C'
    # Loads the argument at index 5 onto the evaluation stack.
    LDARG5 = b'\x7D'
    # Loads the argument at index 6 onto the evaluation stack.
    LDARG6 = b'\x7E'
    # Loads the argument at a specified index onto the evaluation stack. The index is represented as a 1-byte
    # unsigned integer.
    LDARG = b'\x7F'
    # Stores the value on top of the evaluation stack in the argument slot at index 0.
    STARG0 = b'\x80'
    # Stores the value on top of the evaluation stack in the argument slot at index 1.
    STARG1 = b'\x81'
    # Stores the value on top of the evaluation stack in the argument slot at index 2.
    STARG2 = b'\x82'
    # Stores the value on top of the evaluation stack in the argument slot at index 3.
    STARG3 = b'\x83'
    # Stores the value on top of the evaluation stack in the argument slot at index 4.
    STARG4 = b'\x84'
    # Stores the value on top of the evaluation stack in the argument slot at index 5.
    STARG5 = b'\x85'
    # Stores the value on top of the evaluation stack in the argument slot at index 6.
    STARG6 = b'\x86'
    # Stores the value on top of the evaluation stack in the argument slot at a specified index. The index is
    # represented as a 1-byte unsigned integer.
    STARG = b'\x87'

    # endregion

    # region Splice

    # Creates a new Buffer and pushes it onto the stack.
    NEWBUFFER = b'\x88'
    # Copies a range of bytes from one Buffer to another.
    MEMCPY = b'\x89'
    # Concatenates two strings.
    CAT = b'\x8B'
    # Returns a section of a string.
    SUBSTR = b'\x8C'
    # Keeps only characters left of the specified point in a string.
    LEFT = b'\x8D'
    # Keeps only characters right of the specified point in a string.
    RIGHT = b'\x8E'

    # endregion

    # region Bitwise logic

    # Flips all of the bits in the input.
    INVERT = b'\x90'
    # Boolean and between each bit in the inputs.
    AND = b'\x91'
    # Boolean or between each bit in the inputs.
    OR = b'\x92'
    # Boolean exclusive or between each bit in the inputs.
    XOR = b'\x93'
    # Returns 1 if the inputs are exactly equal, 0 otherwise.
    EQUAL = b'\x97'
    # Returns 1 if the inputs are not equal, 0 otherwise.
    NOTEQUAL = b'\x98'

    # endregion

    # region Arithmetic

    # Puts the sign of top stack item on top of the main stack. If value is negative, put -1; if positive,
    # put 1; if value is zero, put 0.
    SIGN = b'\x99'
    # The input is made positive.
    ABS = b'\x9A'
    # The sign of the input is flipped.
    NEGATE = b'\x9B'
    # 1 is added to the input.
    INC = b'\x9C'
    # 1 is subtracted from the input.
    DEC = b'\x9D'
    # a is added to b.
    ADD = b'\x9E'
    # b is subtracted from a.
    SUB = b'\x9F'
    # a is multiplied by b.
    MUL = b'\xA0'
    # a is divided by b.
    DIV = b'\xA1'
    # Returns the remainder after dividing a by b.
    MOD = b'\xA2'
    # The result of raising value to the exponent power.
    POW = b'\xA3'
    # Returns the square root of a specified number.
    SQRT = b'\xA4'
    # Performs modulus division on a number multiplied by another number.
    MODMUL = b'\xA5'
    # Performs modulus division on a number raised to the power of another number.
    # If the exponent is -1, it will have the calculation of the modular inverse.
    MODPOW = b'\xA6'
    # Shifts a left b bits, preserving sign.
    SHL = b'\xA8'
    # Shifts a right b bits, preserving sign.
    SHR = b'\xA9'
    # If the input is 0 or 1, it is flipped. Otherwise the output will be 0.
    NOT = b'\xAA'
    # If both a and b are not 0, the output is 1. Otherwise 0.
    BOOLAND = b'\xAB'
    # If a or b is not 0, the output is 1. Otherwise 0.
    BOOLOR = b'\xAC'
    # Returns 0 if the input is 0. 1 otherwise.
    NZ = b'\xB1'
    # Returns 1 if the numbers are equal, 0 otherwise.
    NUMEQUAL = b'\xB3'
    # Returns 1 if the numbers are not equal, 0 otherwise.
    NUMNOTEQUAL = b'\xB4'
    # Returns 1 if a is less than b, 0 otherwise.
    LT = b'\xB5'
    # Returns 1 if a is less than or equal to b, 0 otherwise.
    LE = b'\xB6'
    # Returns 1 if a is greater than b, 0 otherwise.
    GT = b'\xB7'
    # Returns 1 if a is greater than or equal to b, 0 otherwise.
    GE = b'\xB8'
    # Returns the smaller of a and b.
    MIN = b'\xB9'
    # Returns the larger of a and b.
    MAX = b'\xBA'
    # Returns 1 if x is within the specified range (left-inclusive), 0 otherwise.
    WITHIN = b'\xBB'

    # endregion

    # region Compound-type

    # A value n is taken from top of main stack. The next n*2 items on main stack are removed, put inside n-sized map
    # and this map is put on top of the main stack.
    PACKMAP = b'\xBE'
    # A value n is taken from top of main stack. The next n items on main stack are removed, put inside n-sized
    # struct and this struct is put on top of the main stack.
    PACKSTRUCT = b'\xBF'
    # A value n is taken from top of main stack. The next n items on main stack are removed, put inside n-sized array
    # and this array is put on top of the main stack.
    PACK = b'\xC0'
    # A collection is removed from top of the main stack. Its elements are put on top of the main stack (in reverse order) and the collection size is also put on main stack.
    UNPACK = b'\xC1'
    # An empty array (with size 0) is put on top of the main stack.
    NEWARRAY0 = b'\xC2'
    # A value n is taken from top of main stack. A null-filled array with size n is put on top of the main stack.
    NEWARRAY = b'\xC3'
    # A value n is taken from top of main stack. An array of type T with size n is put on top of the main stack.
    NEWARRAY_T = b'\xC4'
    # An empty struct (with size 0) is put on top of the main stack.
    NEWSTRUCT0 = b'\xC5'
    # A value n is taken from top of main stack. A zero-filled struct with size n is put on top of the main stack.
    NEWSTRUCT = b'\xC6'
    # A Map is created and put on top of the main stack.
    NEWMAP = b'\xC8'
    # An array is removed from top of the main stack. Its size is put on top of the main stack.
    SIZE = b'\xCA'
    # An input index n (or key) and an array (or map) are removed from the top of the main stack. Puts True on top of
    # main stack if array[n] (or map[n]) exist, and False otherwise.
    HASKEY = b'\xCB'
    # A map is taken from top of the main stack. The keys of this map are put on top of the main stack.
    KEYS = b'\xCC'
    # A map is taken from top of the main stack. The values of this map are put on top of the main stack.
    VALUES = b'\xCD'
    # An input index n (or key) and an array (or map) are taken from main stack. Element array[n] (or map[n]) is put
    # on top of the main stack.
    PICKITEM = b'\xCE'
    # The item on top of main stack is removed and appended to the second item on top of the main stack.
    APPEND = b'\xCF'
    # A value v, index n (or key) and an array (or map) are taken from main stack. Attribution array[n]=v
    # (or map[n]=v) is performed.
    SETITEM = b'\xD0'
    # An array is removed from the top of the main stack and its elements are reversed.
    REVERSEITEMS = b'\xD1'
    # An input index n (or key) and an array (or map) are removed from the top of the main stack. Element array[n]
    # (or map[n]) is removed.
    REMOVE = b'\xD2'
    # Remove all the items from the compound-type.
    CLEARITEMS = b'\xD3'
    # Remove the last element from an array, and push it onto the stack.
    POPITEM = b'\xD4'

    # endregion

    # region Types

    # Returns true if the input is null;
    ISNULL = b'\xD8'
    # Returns true if the top item of the stack is of the specified type;
    ISTYPE = b'\xD9'
    # Returns true if the input is null;
    CONVERT = b'\xDB'

    # endregion

    def __repr__(self) -> str:
        return str(self)



from boa3.builtin.interop.contract.contractmanifest import ContractManifest
from boa3.builtin.type import UInt160


class Contract:
    """
    Represents a contract that can be invoked.

    :ivar id: the serial number of the contract
    :vartype id: int
    :ivar update_counter: the number of times the contract was updated
    :vartype update_counter: int
    :ivar hash: the hash of the contract
    :vartype hash: UInt160
    :ivar nef: the serialized Neo Executable Format (NEF) object holding of the smart contract code and compiler
        information
    :vartype nef: bytes
    :ivar manifest: the manifest of the contract
    :vartype manifest: ContractManifest
    """

    def __init__(self):
        self.id: int = 0
        self.update_counter: int = 0
        self.hash: UInt160 = UInt160()
        self.nef: bytes = bytes()
        self.manifest: ContractManifest = ContractManifest()



from __future__ import annotations

import ast
from enum import Enum, auto
from typing import Optional, Union

from boa3.model.operation.operation import IOperation
from boa3.model.operation.operator import Operator


class Operation(Enum):
    Add = auto()
    Sub = auto()
    Mult = auto()
    Div = auto()
    FloorDiv = auto()
    Mod = auto()

    @property
    def is_symmetric(self) -> bool:
        return self._symmetric

    @classmethod
    def get_operation(cls, op: Union[ast.operator, IOperation]) -> Optional[Operation]:
        op = op.operator if isinstance(op, IOperation) else op
        if op is Operator.Plus or isinstance(op, (ast.Add, ast.UAdd)):
            return cls.Add
        elif op is Operator.Minus or isinstance(op, (ast.Sub, ast.USub)):
            return cls.Sub
        elif op is Operator.Mult or isinstance(op, ast.Mult):
            return cls.Mult
        elif op is Operator.Div or isinstance(op, ast.Div):
            return cls.Div
        elif op is Operator.IntDiv or isinstance(op, ast.FloorDiv):
            return cls.FloorDiv
        elif op is Operator.Mod or isinstance(op, ast.Mod):
            return cls.Mod



from __future__ import annotations

from typing import Any, Dict, Optional

from boa3.constants import NEO_SCRIPT
from boa3.model.method import Method
from boa3.model.property import Property
from boa3.model.type.classes.classarraytype import ClassArrayType
from boa3.model.variable import Variable


class NeoClass(ClassArrayType):
    """
    A class used to represent NEO native contract
    """

    def __init__(self):
        super().__init__('NEO')

        self._variables: Dict[str, Variable] = {}
        self._class_methods: Dict[str, Method] = {}
        self._constructor: Optional[Method] = None

    @property
    def instance_variables(self) -> Dict[str, Variable]:
        return self._variables.copy()

    @property
    def class_variables(self) -> Dict[str, Variable]:
        return {}

    @property
    def properties(self) -> Dict[str, Property]:
        return {}

    @property
    def static_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def class_methods(self) -> Dict[str, Method]:
        # avoid recursive import
        from boa3.model.builtin.native.nep17_methods import (BalanceOfMethod, DecimalsMethod, SymbolMethod,
                                                             TotalSupplyMethod, TransferMethod)
        from boa3.model.builtin.native.neo_contract_methods import (GetAccountStateMethod, GetAllCandidatesMethod,
                                                                    GetCandidatesMethod, GetCandidateVoteMethod,
                                                                    GetCommitteeMethod, GetGasPerBlockMethod,
                                                                    GetNextBlockValidatorsMethod,
                                                                    RegisterCandidateMethod, UnclaimedGasMethod,
                                                                    UnregisterCandidateMethod, UnVoteMethod, VoteMethod)
        from boa3.model.builtin.contract import NeoAccountStateType

        if len(self._class_methods) == 0:
            self._class_methods = {
                'balanceOf': BalanceOfMethod(NEO_SCRIPT),
                'decimals': DecimalsMethod(NEO_SCRIPT),
                'symbol': SymbolMethod(NEO_SCRIPT),
                'totalSupply': TotalSupplyMethod(NEO_SCRIPT),
                'transfer': TransferMethod(NEO_SCRIPT),

                'get_account_state': GetAccountStateMethod(NeoAccountStateType.build()),
                'get_all_candidates': GetAllCandidatesMethod(),
                'get_candidates': GetCandidatesMethod(),
                'get_candidate_vote': GetCandidateVoteMethod(),
                'get_committee': GetCommitteeMethod(),
                'get_gas_per_block': GetGasPerBlockMethod(),
                'get_next_block_validators': GetNextBlockValidatorsMethod(),
                'register_candidate': RegisterCandidateMethod(),
                'unclaimed_gas': UnclaimedGasMethod(),
                'unregister_candidate': UnregisterCandidateMethod(),
                'un_vote': UnVoteMethod(),
                'vote': VoteMethod(),
            }
        return self._class_methods

    @property
    def instance_methods(self) -> Dict[str, Method]:
        return {}

    def constructor_method(self) -> Optional[Method]:
        return self._constructor

    @classmethod
    def build(cls, value: Any = None) -> NeoClass:
        if value is None or cls._is_type_of(value):
            return _Neo

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, NeoClass)


_Neo = NeoClass()



from typing import Any

from boa3.builtin.interop.crypto import NamedCurve
from boa3.builtin.type import ByteString, ECPoint


class CryptoLib:
    """
    A class used to represent the CryptoLib native contract
    """

    @classmethod
    def murmur32(cls, data: ByteString, seed: int) -> ByteString:
        """
        Computes the hash value for the specified byte array using the murmur32 algorithm.

        :param data: the input to compute the hash code for
        :type data: ByteString
        :param seed: the seed of the murmur32 hash function
        :type seed: int
        :return: the hash value
        :rtype: ByteString
        """
        pass

    @classmethod
    def sha256(cls, key: Any) -> bytes:
        """
        Encrypts a key using SHA-256.

        :param key: the key to be encrypted
        :type key: Any
        :return: a byte value that represents the encrypted key
        :rtype: bytes
        """
        pass

    @classmethod
    def ripemd160(cls, key: Any) -> bytes:
        """
        Encrypts a key using RIPEMD-160.

        :param key: the key to be encrypted
        :type key: Any
        :return: a byte value that represents the encrypted key
        :rtype: bytes
        """
        pass

    @classmethod
    def verify_with_ecdsa(cls, message: Any, pubkey: ECPoint, signature: ByteString, curve: NamedCurve) -> bool:
        """
        Using the elliptic curve, it checks if the signature of the any item was originally produced by the public key.

        :param message: the encrypted message
        :type message: Any
        :param pubkey: the public key that might have created the item
        :type pubkey: ECPoint
        :param signature: the signature of the item
        :type signature: bytes
        :param curve: the curve that will be used by the ecdsa
        :type curve: NamedCurve
        :return: a boolean value that represents whether the signature is valid
        :rtype: bool
        """
        pass



import ast
from typing import Any, Dict, Iterable, List, Sized

from boa3.model import set_internal_call
from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.builtin.interop.storage import FindOptionsType
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class StorageFindMethod(InteropMethod):

    def __init__(self, find_options_type: FindOptionsType, prefix_type: IType = None):
        from boa3.model.type.type import Type
        from boa3.model.builtin.interop.storage.storagecontext.storagecontexttype import StorageContextType

        identifier = 'find'
        syscall = 'System.Storage.Find'
        context_type = StorageContextType.build()

        if prefix_type is None:
            from boa3.model.type.primitive.bytestringtype import ByteStringType
            prefix_type = ByteStringType.build()

        args: Dict[str, Variable] = {'prefix': Variable(prefix_type),
                                     'context': Variable(context_type),
                                     'options': Variable(find_options_type)}

        from boa3.model.builtin.interop.iterator import IteratorType
        return_type = IteratorType.build(Type.dict.build([prefix_type,  # return an Iterator[prefix, bytes]
                                                          Type.bytes]))

        from boa3.model.builtin.interop.storage.storagegetcontextmethod import StorageGetContextMethod
        default_id = StorageGetContextMethod(context_type).identifier
        context_default = set_internal_call(ast.parse("{0}()".format(default_id)
                                                      ).body[0].value)
        options_default = set_internal_call(ast.parse("{0}.{1}".format(find_options_type.identifier,
                                                                       find_options_type.default_value.name)
                                                      ).body[0].value)

        defaults = [context_default, options_default]

        super().__init__(identifier, syscall, args, defaults=defaults, return_type=return_type)

    @property
    def identifier(self) -> str:
        return '-{0}_{1}'.format(self._identifier, self.prefix_arg.type.identifier)

    @property
    def generation_order(self) -> List[int]:
        """
        Gets the indexes order that need to be used during code generation.
        If the order for generation is the same as inputted in code, returns reversed(range(0,len_args))

        :return: Index order for code generation
        """
        indexes = super().generation_order
        context_index = list(self.args).index('context')

        if indexes[-1] != context_index:
            # context must be the last generated argument
            indexes.remove(context_index)
            indexes.append(context_index)

        return indexes

    @property
    def prefix_arg(self) -> Variable:
        return self.args['prefix']

    @property
    def options_arg(self) -> Variable:
        return self.args['options']

    def build(self, value: Any) -> IBuiltinMethod:
        exp: List[IExpression] = []
        if isinstance(value, Sized):
            if len(value) > 1 or not isinstance(value, Iterable):
                return self
            exp = [exp if isinstance(exp, IExpression) else Variable(exp)
                   for exp in value if isinstance(exp, (IExpression, IType))]

        elif isinstance(exp, (IExpression, IType)):
            exp = [value if isinstance(value, IExpression) else Variable(value)]
        else:
            return self

        if not self.validate_parameters(*exp):
            return self

        method = self
        prefix_type: IType = exp[0].type
        if type(method.prefix_arg.type) != type(prefix_type):
            method = StorageFindMethod(self.options_arg.type, prefix_type)
        return method



from __future__ import annotations

from typing import Any, Dict, Optional, Set


class ScopeValue:
    def __init__(self):
        self._values: Dict[str, Any] = {}
        self._assigned_variables: Set[str] = set()
        self._parent_scope: Optional[ScopeValue] = None

    def new_scope(self) -> ScopeValue:
        scope = ScopeValue()
        scope._values = self._values.copy()
        scope._assigned_variables = self._assigned_variables.copy()
        scope._parent_scope = self
        return scope

    def previous_scope(self) -> Optional[ScopeValue]:
        return self._parent_scope

    def update_values(self, *scopes, is_loop_scope: bool = False):
        other_scopes = [scope for scope in scopes if isinstance(scope, ScopeValue) and scope._parent_scope == self]
        if len(other_scopes) == 0:
            return

        first_scope = other_scopes[0]
        common_keys: Set[str] = set(first_scope._values)
        common_assigns: Set[str] = first_scope._assigned_variables

        for scope in other_scopes[1:]:
            common_keys = common_keys.intersection(set(scope._values))
            common_assigns = common_assigns.intersection(scope._assigned_variables)

        new_values = {}
        different_values = []

        for key in common_keys:
            values = []
            for scope in other_scopes:
                if scope[key] not in values:
                    values.append(scope[key])

            if len(values) == 1:
                new_values[key] = values.pop()
            else:
                different_values.append(key)

        # if there are variables assigned with the same value in all scopes, include this value to the outer scope
        if not is_loop_scope:
            self._values = new_values
        else:
            self._values = {key: value for key, value in new_values.items()
                            if key in self._values and self._values[key] == value}
        self._assigned_variables = self._assigned_variables.union(common_assigns)

        # if there are different value depending on the scope, there's no way to guarantee the value in the actual
        # scope anymore, so remove it from the dictionary
        for key in different_values:
            if key in self._values:
                self._values.pop(key)

    def reset(self):
        self._values.clear()
        self._assigned_variables.clear()

    def assign(self, key: str):
        self._assigned_variables.add(key)

    def remove(self, key: str):
        if key in self._values:
            self._values.pop(key)

    def __contains__(self, item: str) -> bool:
        return item in self._values

    def __getitem__(self, key: str) -> Any:
        if key in self._values:
            return self._values[key]
        else:
            return None

    def __setitem__(self, key: str, value: Any):
        self._values[key] = value


class UndefinedType:
    def __init__(self):
        pass

    @property
    def identifier(self) -> str:
        return 'undefined'


Undefined = UndefinedType()



from typing import Any

from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.StackItem import StackItemType
from boa3.neo3.core.serialization import BinaryReader


def deserialize_binary(reader: BinaryReader) -> Any:
    deserialized_items = []
    underserialized = 1

    while underserialized > 0:
        stack_type = reader.read_byte()
        if stack_type == StackItemType.Any:
            deserialized_items.append(None)
        elif stack_type == StackItemType.Boolean:
            deserialized_items.append(reader.read_bool())
        elif stack_type == StackItemType.Integer:
            value_in_bytes = reader.read_var_bytes()
            deserialized_items.append(Integer.from_bytes(value_in_bytes))
        elif stack_type in (StackItemType.ByteString, StackItemType.Buffer):
            deserialized_items.append(reader.read_var_bytes())
        elif stack_type in (StackItemType.Array, StackItemType.Struct):
            count = reader.read_var_int()
            deserialized_items.append(list(range(count)))
            underserialized += count
        elif stack_type == StackItemType.Map:
            count = reader.read_var_int()
            deserialized_items.append({key: None for key in range(count)})
            underserialized += count * 2
        else:
            raise ValueError

        underserialized -= 1

    stack_temp = []
    while len(deserialized_items) > 0:
        item = deserialized_items.pop()
        if isinstance(item, list):
            new_item = []
            for _ in range(len(item)):
                new_item.append(stack_temp.pop())
            item = new_item
        elif isinstance(item, dict):
            new_item = {}
            for _ in range(0, len(item), 2):
                key = stack_temp.pop()
                value = stack_temp.pop()
                new_item[key] = value
            item = new_item
        stack_temp.append(item)

    return stack_temp.pop()



from typing import Any

from boa3.model.type.itype import IType
from boa3.model.type.primitive.ibytestringtype import IByteStringType
from boa3.neo.vm.type.AbiType import AbiType


class BytesType(IByteStringType):
    """
    A class used to represent Python bytes type
    """

    def __init__(self):
        identifier = 'bytes'
        from boa3.model.type.primitive.inttype import IntType
        values_type = [IntType()]
        super().__init__(identifier, values_type)

    @property
    def abi_type(self) -> AbiType:
        return AbiType.ByteArray

    @property
    def default_value(self) -> Any:
        return bytes()

    @classmethod
    def build(cls, value: Any) -> IType:
        from boa3.model.type.type import Type
        return Type.bytes

    @classmethod
    def build_collection(cls, *value_type: IType):
        return cls.build(value_type)

    @classmethod
    def _is_type_of(cls, value: Any):
        return type(value) is bytes or isinstance(value, BytesType)

    def _init_class_symbols(self):
        super()._init_class_symbols()

        from boa3.model.builtin.builtin import Builtin

        instance_methods = [Builtin.ConvertToBool,
                            Builtin.ConvertToInt,
                            Builtin.ConvertToStr,
                            ]

        for instance_method in instance_methods:
            self._instance_methods[instance_method.raw_identifier] = instance_method.build(self)



import ast
from typing import Any, Dict, List, Optional, Tuple, Union

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ByteArrayMethod(IBuiltinMethod):

    def __init__(self, argument_type: Union[IType, Dict] = None):
        identifier = 'bytearray'
        from boa3.model.type.type import Type

        if isinstance(argument_type, Dict):
            args = argument_type
            defaults = []
        else:
            from boa3.model.type.type import Type
            if argument_type is None or not self.validate_parameters(argument_type):
                argument_type = Type.none

            args: Dict[str, Variable] = {'object': Variable(argument_type)}
            object_default = ast.parse(f"{Type.int.default_value}"
                                       ).body[0].value
            defaults = [object_default]

        super().__init__(identifier, args, defaults=defaults, return_type=Type.bytearray)

    @property
    def _arg_object(self) -> Variable:
        return self.args['object']

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type
        if self._arg_object.type is Type.none:
            return self._identifier

        if self._args_on_stack == 2:
            return '-{0}_with_enconding'.format(self._identifier)

        return '-{0}_from_{1}'.format(self._identifier, self._arg_object.type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) > 1:
            return False
        if len(params) == 0:
            return True

        from boa3.model.type.itype import IType
        if not isinstance(params[0], (IExpression, IType)):
            return False

        param_type: IType = params[0].type if isinstance(params[0], IExpression) else params[0]
        from boa3.model.type.type import Type

        if Type.int.is_type_of(param_type):
            return True

        return (isinstance(param_type, SequenceType)
                and (param_type is Type.str
                     or isinstance(param_type.value_type, type(Type.int))
                     ))

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.StackItem import StackItemType
        from boa3.model.type.type import Type

        if self._arg_object.type is Type.int:
            return [(Opcode.NEWBUFFER, b'')]
        else:
            return [(Opcode.CONVERT, StackItemType.Buffer)]

    @property
    def is_supported(self) -> bool:
        # TODO: change when building bytearray from int iterators are implemented
        from boa3.model.type.type import Type
        return self._arg_object.type in (Type.bytes, Type.str, Type.int)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self._arg_object.type):
            return self
        if isinstance(value, list):
            if len(value) == 2:
                from boa3.model.builtin.method.bytearrayencodingmethod import ByteArrayEncodingMethod
                return ByteArrayEncodingMethod()

            value = value[0] if len(value) > 0 else None
        if self.validate_parameters(value):
            return ByteArrayMethod(value)
        return super().build(value)



from typing import Any

from boa3 import constants
from boa3.model.type.itype import IType
from boa3.model.type.primitive.inttype import IntType
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.StackItem import StackItemType


class BoolType(IntType):
    """
    A class used to represent Python bool type
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'bool'

    @property
    def default_value(self) -> Any:
        return bool()

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Boolean

    @property
    def stack_item(self) -> StackItemType:
        return StackItemType.Boolean

    def _init_class_symbols(self):
        super()._init_class_symbols()

        from boa3.model.builtin.builtin import Builtin
        self._instance_methods[constants.INIT_METHOD_ID] = Builtin.Bool

    @classmethod
    def build(cls, value: Any) -> IType:
        if cls._is_type_of(value):
            from boa3.model.type.type import Type
            return Type.bool

    @classmethod
    def _is_type_of(cls, value: Any):
        return type(value) in [bool, BoolType]



import ast
from abc import ABC
from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.builtincallable import IBuiltinCallable
from boa3.model.event import Event
from boa3.model.variable import Variable


class IBuiltinEvent(IBuiltinCallable, Event, ABC):
    def __init__(self, identifier: str, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None,
                 vararg: Optional[Tuple[str, Variable]] = None,
                 kwargs: Optional[Dict[str, Variable]] = None):
        from boa3.model.type.type import Type
        super().__init__(identifier, args, vararg, kwargs, defaults, Type.none)

        # constructor of IBuiltinCallable and Event classes are conflicting
        self._identifier = identifier
        self.name = identifier
        self.args = args if args is not None else {}
        self.defaults = defaults if defaults is not None else []



import ast
from typing import Dict, List, Optional, Tuple

from boa3.model.callable import Callable
from boa3.model.debuginstruction import DebugInstruction
from boa3.model.symbol import ISymbol
from boa3.model.type.classes.classtype import ClassType
from boa3.model.type.type import IType, Type
from boa3.model.variable import Variable


class Method(Callable):
    """
    A class used to represent a function or a class method

    :ivar args: a dictionary that maps each arg with its name. Empty by default.
    :ivar locals: a dictionary that maps each local variable with its name. Empty by default.
    :ivar imported_symbols: a dictionary that maps each imported symbol with its name. Empty by default.
    :ivar is_public: a boolean value that specifies if the method is public. False by default.
    :ivar return_type: the return type of the method. None by default.
    """

    def __init__(self, args: Dict[str, Variable] = None,
                 vararg: Optional[Tuple[str, Variable]] = None,
                 kwargs: Optional[Dict[str, Variable]] = None,
                 defaults: List[ast.AST] = None,
                 return_type: IType = Type.none, is_public: bool = False,
                 decorators: List[Callable] = None,
                 is_init: bool = False,
                 external_name: str = None,
                 is_safe: bool = False,
                 origin_node: Optional[ast.AST] = None):
        super().__init__(args, vararg, kwargs, defaults, return_type, is_public, decorators, external_name, is_safe, origin_node)

        self.imported_symbols = {}
        self._symbols = {}
        self.defined_by_entry = True
        self.is_init = is_init
        self.locals: Dict[str, Variable] = {}

        if is_init and self.has_cls_or_self:
            self.return_type = list(self.args.values())[0].type

        self._debug_map: List[DebugInstruction] = []
        self.origin_class: Optional[ClassType] = None

    @property
    def shadowing_name(self) -> str:
        return 'method'

    def include_variable(self, var_id: str, var: Variable, is_global: bool = False):
        """
        Includes a variable into the list of locals

        :param var_id: variable identifier
        :param var: variable to be included
        :param is_global: whether the variable is declared outside the method
        """
        if var_id not in self.symbols:
            if is_global:
                self._symbols[var_id] = var
            else:
                self.locals[var_id] = var

    @property
    def symbols(self) -> Dict[str, Variable]:
        """
        Gets all the symbols in the method

        :return: a dictionary that maps each symbol in the module with its name
        """
        symbols = self._symbols.copy()
        symbols.update(self.imported_symbols)
        symbols.update(self.args)
        symbols.update(self.locals)
        return symbols

    def include_symbol(self, symbol_id: str, symbol: ISymbol, is_global: bool = False):
        """
        Includes a method into the scope of the module

        :param symbol_id: method identifier
        :param symbol: method to be included
        :param is_global: whether the variable is declared outside the method
        """
        if symbol_id not in self.symbols:
            if isinstance(symbol, Variable):
                self.include_variable(symbol_id, symbol, is_global)
            else:
                self.imported_symbols[symbol_id] = symbol

    @property
    def origin(self) -> ast.AST:
        """
        Returns the method origin ast node.

        :return: the ast node that describes this method. None if it is not from a ast.
        """
        return self._origin_node

    def debug_map(self) -> List[DebugInstruction]:
        """
        Returns a list with the debug information of each mapped Python instruction inside this method
        """
        from boa3.compiler.codegenerator.vmcodemapping import VMCodeMapping
        return sorted(self._debug_map, key=lambda instr: VMCodeMapping.instance().get_start_address(instr.code))

    def include_instruction(self, instr_info: DebugInstruction):
        """
        Includes a new instruction in the debug info

        :param instr_info: debug information from the new instruction
        """
        if not any((info.start_line == instr_info.start_line and info.start_col == instr_info.start_col
                    for info in self._debug_map)):
            existing_instr_info: Optional[DebugInstruction] = \
                next((info for info in self._debug_map if info.code == instr_info.code), None)
            if existing_instr_info is not None:
                self._debug_map.remove(existing_instr_info)
            self._debug_map.append(instr_info)

    def remove_instruction(self, start_line: int, start_col: int):
        """
        Removes a instruction from the debug info at the given position if it exists

        :param start_line: instruction's first line
        :param start_col: instruction's beginning offset in the first line
        """
        instruction = next((info for info in self._debug_map
                            if info.start_line == start_line and info.start_col == start_col),
                           None)
        if instruction is not None:
            self._debug_map.remove(instruction)

    def args_to_be_generated(self) -> List[int]:
        """
        Gets the indexes of the arguments that must be generated.
        If method has `self` arg, it must be in this list.

        By default, all the arguments are generated.

        :return: A list with the indexes of the arguments that must be generated
        """
        return list(range(len(self.args)))



from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class PowMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'pow'
        args: Dict[str, Variable] = {
            'base': Variable(Type.int),
            'exponent': Variable(Type.int),
        }
        super().__init__(identifier, args, return_type=Type.int)

    @property
    def requires_reordering(self) -> bool:
        return True

    def reorder(self, arguments: list):
        # swap base and exponent  positions
        arguments[0], arguments[1] = arguments[1], arguments[0]

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.POW, b'')]

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return



from typing import Any, Dict, Tuple, Union

from boa3.neo3.core.types import UInt160
from test_runner import utils


class NeoInvoke:
    def __init__(self, contract_id: Union[str, UInt160], operation: str, *args: Any):
        self._contract_id = str(contract_id)
        self._operation = operation
        self._args = tuple(utils.value_to_parameter(x) for x in args)
        self._invoker = 'genesis'  # TODO: allow to change to any account

    @property
    def contract(self) -> str:
        return self._contract_id

    @property
    def operation(self) -> str:
        return self._operation

    @property
    def args(self) -> Tuple[str]:
        import json
        return tuple((json.dumps(arg, separators=(',', ':')) for arg in self._args))

    @property
    def invoker(self) -> str:
        return self._invoker

    def to_json(self) -> Dict[str, Any]:
        return {
            'contract': self._contract_id,
            'operation': self._operation,
            'args': list(self._args)
        }



import ast
from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.classmethod.indexmethod import IndexMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String


class IndexSequenceMethod(IndexMethod):

    def __init__(self, sequence_type: Optional[SequenceType] = None, arg_value: Optional[IType] = None):
        from boa3.model.type.type import Type
        if not isinstance(sequence_type, SequenceType):
            sequence_type = Type.sequence
        if not isinstance(arg_value, IType):
            arg_value = Type.any

        args: Dict[str, Variable] = {
            'self': Variable(sequence_type),
            'x': Variable(arg_value),
            'start': Variable(Type.int),
            'end': Variable(Type.int),
        }

        start_default = ast.parse("{0}".format(0)
                                  ).body[0].value
        end_default = ast.parse("-1").body[0].value.operand
        end_default.n = -1

        super().__init__(args, defaults=[start_default, end_default])

    def validate_parameters(self, *params: IExpression) -> bool:
        if 2 <= len(params) <= 4:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.model.type.itype import IType
        sequence_type: IType = params[0].type

        if not isinstance(sequence_type, SequenceType):
            return False
        from boa3.model.type.collection.sequence.mutable.listtype import ListType
        from boa3.model.type.collection.sequence.tupletype import TupleType
        from boa3.model.type.collection.sequence.rangetype import RangeType
        if not isinstance(sequence_type, (ListType, TupleType, RangeType)):
            return False
        return True

    @property
    def exception_message(self) -> str:
        return "x not in sequence"

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count

        jmp_place_holder = (Opcode.JMP, b'\x01')
        message = String(self.exception_message).to_bytes()

        # receives: end, start, x, sequence
        verify_negative_index = [           # verifies if index in a negative value
            (Opcode.DUP, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder                # if index >= 0, jump to verify_big_end or verify_big_start
        ]

        fix_negative_end = [                # gets the correspondent positive value of end
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.ADD, b''),
            (Opcode.INC, b''),              # end = end + len(sequence) + 1
            (Opcode.DUP, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder                # if end is not negative anymore, start verifying start
        ]

        fix_still_negative_index = [        # if index is still negative, consider it 0 then
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),            # end = 0
        ]

        jmp_fix_negative_index = Opcode.get_jump_and_data(Opcode.JMPGT, get_bytes_count(fix_negative_end +
                                                                                        fix_still_negative_index), True)
        verify_negative_index[-1] = jmp_fix_negative_index

        verify_big_end = [                  # verify if end is bigger then len(sequence)
            (Opcode.DUP, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            jmp_place_holder                # if end <= len(sequence), start verifying start
        ]

        fix_big_end = [                     # consider end as len(sequence)
            (Opcode.DROP, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),             # end = len(sequence)
        ]

        jmp_other_verifies = Opcode.get_jump_and_data(Opcode.JMPGT, get_bytes_count(fix_still_negative_index +
                                                                                    verify_big_end +
                                                                                    fix_big_end), True)
        fix_negative_end[-1] = jmp_other_verifies

        jmp_fix_big_index = Opcode.get_jump_and_data(Opcode.JMPLE, get_bytes_count(fix_big_end), True)
        verify_big_end[-1] = jmp_fix_big_index

        verify_and_fix_end = [              # collection of Opcodes regarding verifying and fixing end index
            (Opcode.REVERSE4, b''),
        ]
        verify_and_fix_end.extend(verify_negative_index)
        verify_and_fix_end.extend(fix_negative_end)
        verify_and_fix_end.extend(fix_still_negative_index)
        verify_and_fix_end.extend(verify_big_end)
        verify_and_fix_end.extend(fix_big_end)

        verify_and_fix_start = [            # collection of Opcodes regarding verifying and fixing start index
            (Opcode.SWAP, b'')
        ]
        verify_and_fix_start.extend(verify_negative_index)
        verify_and_fix_start.extend(fix_negative_end)
        verify_and_fix_start.extend(fix_still_negative_index)
        verify_and_fix_start.extend(verify_big_end)
        verify_and_fix_start.extend(fix_big_end)

        verify_while = [                    # verify if already went through all items on the sequence[start:end]
            (Opcode.OVER, b''),             # index = start
            (Opcode.OVER, b''),
            jmp_place_holder                # if index <= start, jump to not_inside_sequence
        ]

        compare_item = [                    # verifies if x is in sequence
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.NUMEQUAL, b''),         # found_x = sequence[index] == x
            jmp_place_holder                # if found_x, jump to return_index
        ]

        not_found = [                       # increments index and goes back to verify again
            (Opcode.INC, b''),              # index++
            # jump to verify_while
        ]

        jmp_back_to_verify = Opcode.get_jump_and_data(Opcode.JMP, -get_bytes_count(verify_while +
                                                                                   compare_item +
                                                                                   not_found), True)
        not_found.append(jmp_back_to_verify)

        jmp_to_error = Opcode.get_jump_and_data(Opcode.JMPLE, get_bytes_count(compare_item +
                                                                              not_found), True)
        verify_while[-1] = jmp_to_error

        not_inside_sequence = [             # send error message saying that x is not in sequence
            (Opcode.PUSHDATA1, Integer(len(message)).to_byte_array(signed=True, min_length=1) + message),
            (Opcode.THROW, b''),
        ]

        jmp_to_return_index = Opcode.get_jump_and_data(Opcode.JMPIF, get_bytes_count(not_found +
                                                                                     not_inside_sequence), True)
        compare_item[-1] = jmp_to_return_index

        return_index = [                    # removes all values in the stack but the index
            (Opcode.NIP, b''),
            (Opcode.NIP, b''),
            (Opcode.NIP, b''),
        ]

        return (
            verify_and_fix_end +
            verify_and_fix_start +
            verify_while +
            compare_item +
            not_found +
            not_inside_sequence +
            return_index
        )



from typing import Dict

from boa3.model.builtin.interop.nativecontract import CryptoLibMethod
from boa3.model.variable import Variable


class Ripemd160Method(CryptoLibMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'ripemd160'
        native_identifier = 'ripemd160'
        args: Dict[str, Variable] = {'key': Variable(Type.any)}
        super().__init__(identifier, native_identifier, args, return_type=Type.bytes)



from typing import Any, List, Sequence

from boa3.builtin.interop.contract.callflagstype import CallFlags
from boa3.builtin.interop.contract.contract import Contract
from boa3.builtin.type import ECPoint, UInt160


def call_contract(script_hash: UInt160, method: str, args: Sequence = (), call_flags: CallFlags = CallFlags.ALL) -> Any:
    """
    Calls a smart contract given the method and the arguments.

    :param script_hash: the target smart contract's script hash
    :type script_hash: UInt160
    :param method: the name of the method to be executed
    :type method: str
    :param args: the specified method's arguments
    :type args: Sequence[Any]
    :param call_flags: the CallFlags to be used to call the contract
    :type call_flags: CallFlags

    :return: the result of the specified method
    :rtype: Any

    :raise Exception: raised if there isn't a valid CallFlags, the script hash is not a valid smart contract or the
        method was not found or the arguments aren't valid to the specified method.
    """
    pass


def create_contract(nef_file: bytes, manifest: bytes, data: Any = None) -> Contract:
    """
    Creates a smart contract given the script and the manifest.

    :param nef_file: the target smart contract's compiled nef
    :type nef_file: bytes
    :param manifest: the manifest.json that describes how the script should behave
    :type manifest: bytes
    :param data: the parameters for the _deploy function
    :type data: Any

    :return: the contract that was created
    :rtype: Contract

    :raise Exception: raised if the nef or the manifest are not a valid smart contract.
    """
    pass


def update_contract(nef_file: bytes, manifest: bytes, data: Any = None):
    """
    Updates the executing smart contract given the script and the manifest.

    :param nef_file: the new smart contract's compiled nef
    :type nef_file: bytes
    :param manifest: the new smart contract's manifest
    :type manifest: bytes
    :param data: the parameters for the _deploy function
    :type data: Any

    :raise Exception: raised if the nef and the manifest are not a valid smart contract or the new contract is the
        same as the old one.
    """
    pass


def destroy_contract():
    """
    Destroy the executing smart contract.
    """
    pass


def get_minimum_deployment_fee() -> int:
    """
    Gets the minimum fee of contract deployment.

    :return: the minimum fee of contract deployment
    """
    pass


def get_call_flags() -> CallFlags:
    """
    Gets the CallFlags in the current context.
    """
    pass


def create_standard_account(pub_key: ECPoint) -> UInt160:
    """
    Calculates the script hash from a public key.

    :param pub_key: the given public key
    :type pub_key: ECPoint

    :return: the corresponding script hash of the public key
    :rtype: UInt160
    """
    pass


def create_multisig_account(m: int, pub_keys: List[ECPoint]) -> UInt160:
    """
    Calculates corresponding multisig account script hash for the given public keys.

    :param m: the minimum number of correct signatures need to be provided in order for the verification to pass.
    :type m: int
    :param pub_keys: the public keys of the account
    :type pub_keys: List[ECPoint]

    :return: the hash of the corresponding account
    :rtype: UInt160
    """
    pass


NEO: UInt160 = UInt160()
"""
NEO's token script hash.

:meta hide-value:
"""

GAS: UInt160 = UInt160()
"""
GAS' token script hash.

:meta hide-value:
"""



from typing import Dict

from boa3.model.builtin.builtinproperty import IBuiltinProperty
from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.variable import Variable


class GetEntryScriptHashMethod(InteropMethod):
    def __init__(self):
        from boa3.model.type.collection.sequence.uint160type import UInt160Type
        identifier = '-get_entry_script_hash'
        syscall = 'System.Runtime.GetEntryScriptHash'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, syscall, args, return_type=UInt160Type.build())


class EntryScriptHashProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'entry_script_hash'
        getter = GetEntryScriptHashMethod()
        super().__init__(identifier, getter)



class NotLoadedException(Exception):
    def __init__(self, empty_script: bool = False):
        self.empty_script = empty_script

    @property
    def message(self) -> str:
        exception_message = ''
        if self.empty_script:
            exception_message = 'An empty script was generated'

        return exception_message



from typing import Dict

from boa3.model.builtin.interop.nativecontract import RoleManagementMethod
from boa3.model.variable import Variable


class GetDesignatedByRoleMethod(RoleManagementMethod):
    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.builtin.interop.role.roletype import RoleType
        from boa3.model.type.collection.sequence.ecpointtype import ECPointType

        identifier = 'get_designated_by_role'
        native_identifier = 'getDesignatedByRole'

        role_type = RoleType.build()
        args: Dict[str, Variable] = {
            'role': Variable(role_type),
            'index': Variable(Type.int)
        }
        super().__init__(identifier, native_identifier, args, return_type=ECPointType.build())



__all__ = ['ClassMethodDecorator',
           'ContractDecorator',
           'DisplayNameDecorator',
           'InstanceMethodDecorator',
           'MetadataDecorator',
           'PropertyDecorator',
           'PublicDecorator',
           'StaticMethodDecorator'
           ]

from boa3.model.builtin.decorator.classmethoddecorator import ClassMethodDecorator
from boa3.model.builtin.decorator.contractdecorator import ContractDecorator
from boa3.model.builtin.decorator.displaynamedecorator import DisplayNameDecorator
from boa3.model.builtin.decorator.instancemethoddecorator import InstanceMethodDecorator
from boa3.model.builtin.decorator.metadatadecorator import MetadataDecorator
from boa3.model.builtin.decorator.propertydecorator import PropertyDecorator
from boa3.model.builtin.decorator.publicdecorator import PublicDecorator
from boa3.model.builtin.decorator.staticmethoddecorator import StaticMethodDecorator



from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.classmethod.countmethod import CountMethod
from boa3.model.expression import IExpression
from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class CountSequenceMethod(CountMethod):

    def __init__(self, sequence_type: Optional[SequenceType] = None, arg_value: Optional[IType] = None):
        from boa3.model.type.type import Type
        if not isinstance(sequence_type, SequenceType):
            sequence_type = Type.sequence
        if not isinstance(arg_value, IType):
            arg_value = Type.any

        args: Dict[str, Variable] = {
            'self': Variable(sequence_type),
            'value': Variable(arg_value)
        }

        self._generic_verification_opcodes = None

        super().__init__(args)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 2:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.model.type.itype import IType
        sequence_type: IType = params[0].type

        if not isinstance(sequence_type, SequenceType):
            return False
        from boa3.model.type.collection.sequence.mutable.listtype import ListType
        from boa3.model.type.collection.sequence.tupletype import TupleType
        from boa3.model.type.collection.sequence.rangetype import RangeType
        if not isinstance(sequence_type, (ListType, TupleType, RangeType)):
            return False
        return True

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count

        jmp_place_holder = (Opcode.JMP, b'\x01')

        # region Sequence logic

        repack_array = [  # recreates an array to not change the original one
            (Opcode.UNPACK, b''),
            (Opcode.PACK, b''),
        ]

        sequence_initialize = [  # initializes variables
            (Opcode.PUSH0, b''),  # count = 0
            (Opcode.OVER, b''),
            (Opcode.SIZE, b''),
            (Opcode.DEC, b''),  # index = len(sequence) - -1
        ]

        sequence_verify_while = [  # verifies if all the elements from the sequence were visited
            (Opcode.DUP, b''),  # would be equivalent to: while (index >= 0)
            (Opcode.SIGN, b''),
            (Opcode.PUSH0, b''),  # if index < 0:
            jmp_place_holder,  # jump to clean the stack
        ]

        sequence_get_element = [  # gets a element from the sequence
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b'')  # element = sequence[index]
        ]

        sequence_equals = [  # verifies if the element and the given value are the same
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.EQUAL, b''),  # if element != value:
            jmp_place_holder  # jump to list_tuple_count_index_dec
        ]

        sequence_count_inc = [  # increment count if element == value
            (Opcode.SWAP, b''),
            (Opcode.INC, b''),  # count++
            (Opcode.SWAP, b''),
        ]

        num_jmp_code = get_bytes_count(sequence_count_inc)
        jmp_to_dec_statement = Opcode.get_jump_and_data(Opcode.JMPIFNOT, num_jmp_code, True)
        sequence_equals[-1] = jmp_to_dec_statement

        list_tuple_count_index_dec = [  # decreases the index
            (Opcode.DEC, b''),  # index--
            # return to the while verification
        ]

        in_depth_verification = self.generic_verification(get_bytes_count(sequence_count_inc),
                                                          get_bytes_count(sequence_get_element + sequence_equals))

        num_jmp_code = -get_bytes_count(list_tuple_count_index_dec + sequence_count_inc + sequence_equals +
                                        sequence_get_element + sequence_verify_while + in_depth_verification)
        jmp_back_to_while_verify_statement = Opcode.get_jump_and_data(Opcode.JMP, num_jmp_code)
        list_tuple_count_index_dec.append(jmp_back_to_while_verify_statement)

        num_jmp_code = get_bytes_count(sequence_get_element + sequence_equals +
                                       sequence_count_inc + list_tuple_count_index_dec + in_depth_verification)
        jmp_to_clean_statement = Opcode.get_jump_and_data(Opcode.JMPLT, num_jmp_code, True)
        sequence_verify_while[-1] = jmp_to_clean_statement

        sequence_clean_stack = [
            (Opcode.DROP, b''),
            (Opcode.REVERSE3, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
        ]

        # endregion

        return (
            repack_array +
            sequence_initialize +
            sequence_verify_while +
            in_depth_verification +
            sequence_get_element +
            sequence_equals +
            sequence_count_inc +
            list_tuple_count_index_dec +
            sequence_clean_stack
        )

    def generic_verification(self, inc_statement_bytes=None,
                             get_equals_statement_bytes=None) -> List[Tuple[Opcode, bytes]]:
        if self._generic_verification_opcodes is None:
            self._generic_verification_opcodes = []

        return self._generic_verification_opcodes



from typing import Any, Dict

from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType
from boa3.model.type.primitive.inttype import IntType


class TriggerType(IntType):
    """
    A class used to represent Neo interop Trigger type
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'TriggerType'

    @property
    def default_value(self) -> Any:
        from boa3.builtin.interop.runtime import TriggerType as Trigger
        return Trigger.ALL

    @classmethod
    def build(cls, value: Any) -> IType:
        if cls._is_type_of(value):
            from boa3.model.builtin.interop.interop import Interop
            return Interop.TriggerType

    @classmethod
    def _is_type_of(cls, value: Any):
        from boa3.builtin.interop.runtime import TriggerType as Trigger
        return isinstance(value, (Trigger, TriggerType))

    @property
    def symbols(self) -> Dict[str, ISymbol]:
        """
        Gets the class symbols of this type

        :return: a dictionary that maps each symbol in the module with its name
        """
        from boa3.builtin.interop.runtime import TriggerType as Trigger
        from boa3.model.variable import Variable

        _symbols = super().symbols
        _symbols.update({name: Variable(self) for name in Trigger.__members__.keys()})

        return _symbols

    def get_value(self, symbol_id) -> Any:
        """
        Gets the literal value of a symbol

        :return: the value if this type has this symbol. None otherwise.
        """
        if symbol_id in self.symbols:
            from boa3.builtin.interop.runtime import TriggerType as Trigger
            return Trigger.__members__[symbol_id]

        return None



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

from boa3.compiler.codegenerator import get_bytes_count
from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.StackItem import StackItemType


class ClassType(IType, ABC):
    """
    An abstract class used to represent Python class
    """

    def __init__(self, identifier: str, decorators: list = None,
                 bases: List[ClassType] = None):
        super().__init__(identifier)

        if decorators is None:
            decorators = []
        else:
            # avoid circular import
            from boa3.model.decorator import IDecorator
            decorators = [decorator for decorator in decorators
                          if isinstance(decorator, IDecorator)]

        self.decorators = decorators

        if not isinstance(bases, list):
            bases = []
        self.bases: List[ClassType] = bases

    @property
    @abstractmethod
    def class_variables(self):
        """
        :rtype: Dict[str, boa3.model.variable.Variable]
        """
        class_vars = {}
        for base in self.bases:
            class_vars.update(base.class_variables)
        return class_vars

    @property
    @abstractmethod
    def instance_variables(self):
        """
        :rtype: Dict[str, boa3.model.variable.Variable]
        """
        instance_vars = {}
        for base in self.bases:
            instance_vars.update(base.instance_variables)
        return instance_vars

    @property
    def variables(self):
        """
        :rtype: Dict[str, boa3.model.variable.Variable]
        """
        variables = self.class_variables.copy()
        variables.update(self.instance_variables)
        return variables

    @property
    def _all_variables(self):
        """
        :rtype: Dict[str, boa3.model.variable.Variable]
        """
        return self.variables

    @property
    @abstractmethod
    def properties(self):
        """
        :rtype: Dict[str, boa3.model.property.Property]
        """
        props = {}
        for base in self.bases:
            props.update(base.properties)
        return props

    @property
    @abstractmethod
    def static_methods(self):
        """
        :rtype: Dict[str, boa3.model.method.Method]
        """
        static_funcs = {}
        for base in self.bases:
            static_funcs.update(base.static_methods)
        return static_funcs

    @property
    @abstractmethod
    def class_methods(self):
        """
        :rtype: Dict[str, boa3.model.method.Method]
        """
        class_funcs = {}
        for base in self.bases:
            class_funcs.update(base.class_methods)
        return class_funcs

    @property
    @abstractmethod
    def instance_methods(self):
        """
        :rtype: Dict[str, boa3.model.method.Method]
        """
        instance_funcs = {}
        for base in self.bases:
            instance_funcs.update(base.instance_methods)
        return instance_funcs

    @abstractmethod
    def constructor_method(self):
        """
        If the class constructor is None, it mustn't allow instantiation of this class

        :rtype: boa3.model.method.Method or None
        """
        pass

    @property
    def symbols(self) -> Dict[str, ISymbol]:
        s = {}
        s.update(self.static_methods)
        s.update(self.class_methods)
        s.update(self.instance_methods)
        s.update(self.variables)
        s.update(self.properties)
        return s

    @property
    def class_symbols(self) -> Dict[str, ISymbol]:
        s: Dict[str, ISymbol] = {}
        s.update(self.class_methods)  # class methods and variables can be accessed both
        s.update(self.class_variables)  # from class name or instance object
        s.update(self.static_methods)
        return s

    @property
    def instance_symbols(self) -> Dict[str, ISymbol]:
        s: Dict[str, ISymbol] = {}
        s.update(self.class_methods)  # class methods and variables can be accessed both
        s.update(self.class_variables)  # from class name or instance object
        s.update(self.instance_methods)
        s.update(self.instance_variables)
        s.update(self.properties)
        return s

    @property
    @abstractmethod
    def abi_type(self) -> AbiType:
        # must be overwritten, classes cannot be mapped to Any
        return super().abi_type

    @property
    @abstractmethod
    def stack_item(self) -> StackItemType:
        # must be overwritten, classes cannot be mapped to Any
        return super().stack_item

    @property
    def is_interface(self) -> bool:
        # TODO: change when other interfaces identifiers are implemented
        from boa3.model.builtin.decorator import ContractDecorator
        return any(isinstance(decorator, ContractDecorator) for decorator in self.decorators)

    def is_instance_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        is_type_opcodes = [
            (Opcode.DUP, b''),  # if is the same internal type
            (Opcode.ISTYPE, self.stack_item)
        ]

        return_false = [
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),  # return False
        ]
        return_false_bytes_size = get_bytes_count(return_false)

        final_instructions = []
        final_jmp = [(Opcode.JMP, Integer(2 + return_false_bytes_size).to_byte_array(min_length=1))]
        validate_type = self._is_instance_inner_opcodes(get_bytes_count(final_jmp))
        if len(validate_type) > 0:
            final_instructions += (validate_type + final_jmp)

            bytes_size = get_bytes_count(final_instructions)
            jmp_opcode, jmp_arg = Opcode.get_jump_and_data(Opcode.JMPIFNOT, bytes_size + 1)

            final_instructions = (is_type_opcodes +
                                  [(jmp_opcode, jmp_arg)] +
                                  final_instructions +
                                  return_false)
        else:
            is_type_opcodes.remove((Opcode.DUP, b''))
            final_instructions = is_type_opcodes

        return final_instructions

    def _is_instance_inner_opcodes(self, jmp_to_if_false: int = 0) -> List[Tuple[Opcode, bytes]]:
        if self.stack_item not in (StackItemType.Array, StackItemType.Struct, StackItemType.Map):
            return []

        variables_count_is_equal = [
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),
            Opcode.get_push_and_data(len(self._all_variables)),
            (Opcode.NUMEQUAL, b''),
        ]

        total_size = jmp_to_if_false
        variables_type_validations = []

        for var_index, (var_id, var) in reversed(list(enumerate(self._all_variables.items()))):

            if var.type.stack_item != StackItemType.Any:

                # validate primitive types only to avoid recursive code
                if self.stack_item == StackItemType.Map:
                    get = Opcode.get_pushdata_and_data(var_id)
                else:
                    get = Opcode.get_push_and_data(var_index)

                new_opcodes = [
                    (Opcode.DUP, b''),
                    get,
                    (Opcode.PICKITEM, b''),
                    (Opcode.ISTYPE, var.type.stack_item)
                ]
                if len(variables_type_validations) == 0:
                    new_opcodes.append((Opcode.NIP, b''))
                elif total_size > 0:
                    new_opcodes.append((Opcode.get_jump_and_data(Opcode.JMPIFNOT, total_size + 1)))

                variables_type_validations = new_opcodes + variables_type_validations
                total_size += get_bytes_count(new_opcodes)

        if total_size > 0:
            if len(variables_type_validations) > 0:
                final_opcode = (Opcode.get_jump_and_data(Opcode.JMPIFNOT, total_size + 1))
            else:
                final_opcode = (Opcode.NIP, b'')
            variables_count_is_equal.append(final_opcode)

        return variables_count_is_equal + variables_type_validations



from typing import Dict, List, Tuple

from boa3.model.builtin.method.listmethod import ListMethod
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ListBytesStringMethod(ListMethod):

    def __init__(self, value: IType = None):
        from boa3.model.type.type import Type

        if value is None:
            value = Type.bytes

        args: Dict[str, Variable] = {
            'value': Variable(value),
        }

        return_type = Type.list.build_collection(value if value is Type.str else Type.int)

        super().__init__(args, return_type)

    @property
    def prepare_for_packing(self) -> List[Tuple[Opcode, bytes]]:

        if self._prepare_for_packing is None:
            from boa3.compiler.codegenerator import get_bytes_count
            from boa3.neo.vm.type.StackItem import StackItemType
            from boa3.model.type.type import Type

            initialize_values = [
                (Opcode.DUP, b''),
                (Opcode.SIZE, b''),
                (Opcode.DEC, b''),
            ]

            loop_for_char_or_byte = [
                (Opcode.OVER, b''),
                (Opcode.OVER, b''),
                (Opcode.PICKITEM, b''),
            ]

            if self._arg_value.type is Type.str:
                loop_for_char_or_byte.append((Opcode.CONVERT, StackItemType.ByteString))

            loop_for_char_or_byte.extend([(Opcode.ROT, b''),
                                          (Opcode.ROT, b''),
                                          (Opcode.DEC, b'')])

            verify_loop_end = [
                (Opcode.DUP, b''),
                (Opcode.SIGN, b''),
                (Opcode.PUSH0, b''),
                # jumps to the beginning of the loop
            ]

            remove_extra_values = [
                (Opcode.DROP, b''),
                (Opcode.SIZE, b''),
            ]

            num_jmp_code = -get_bytes_count(verify_loop_end + loop_for_char_or_byte)
            jmp_to_dec_statement = Opcode.get_jump_and_data(Opcode.JMPGE, num_jmp_code)
            verify_loop_end.append(jmp_to_dec_statement)

            self._prepare_for_packing = (
                initialize_values +
                loop_for_char_or_byte +
                verify_loop_end +
                remove_extra_values
            )

        return super().prepare_for_packing



from typing import Dict, Optional

from boa3.model.builtin.interop.nativecontract import OracleMethod
from boa3.model.variable import Variable


class OracleRequestMethod(OracleMethod):

    def __init__(self):
        from boa3.model.type.type import Type

        identifier = 'request'
        syscall = 'request'
        args: Dict[str, Variable] = {'url': Variable(Type.str),
                                     'request_filter': Variable(Type.union.build([Type.str,
                                                                                  Type.none])),
                                     'callback': Variable(Type.str),
                                     'user_data': Variable(Type.any),
                                     'gas_for_response': Variable(Type.int)}

        super().__init__(identifier, syscall, args, return_type=Type.none)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None



import ast
from typing import Dict

from boa3.analyser.importanalyser import ImportAnalyser
from boa3.model.builtin.builtincallable import IBuiltinCallable
from boa3.model.method import Method
from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class Import(ISymbol):
    """
    A class used to represent an imported package

    :ivar variables: a dictionary that maps each variable with its name. Empty by default.
    :ivar methods: a dictionary that maps each method with its name. Empty by default.
    :ivar types: a dictionary that maps each type with its name. Empty by default.
    """

    def __init__(self, origin: str, syntax_tree: ast.AST, import_analyser: ImportAnalyser,
                 imported_symbols: Dict[str, ISymbol] = None):
        if imported_symbols is None:
            symbols = import_analyser.symbols
        else:
            symbols = import_analyser.export_symbols(list(imported_symbols.keys()))

        self.variables = {var_id: var for var_id, var in symbols.items() if isinstance(var, Variable)}
        self.methods = {fun_id: fun for fun_id, fun in symbols.items() if isinstance(fun, Method)}
        self.types = {type_id: tpe for type_id, tpe in symbols.items() if isinstance(tpe, IType)}
        self.imports = {alias: imprt for alias, imprt in symbols.items() if isinstance(imprt, Import)}
        self._other_symbols = {alias: symbol for alias, symbol in symbols.items()
                               if not isinstance(symbol, (Variable, Method, IType, Import))}

        self._symbols_not_imported = {alias: symbol for alias, symbol in import_analyser.symbols.items()
                                      if alias not in symbols}

        for method in self.methods.values():
            if not isinstance(method, IBuiltinCallable) and hasattr(method, 'defined_by_entry'):
                # methods imported are treated as methods defined in the entry file
                method.defined_by_entry = True

        self.analyser = import_analyser.analyser
        self.origin: str = origin
        self.ast: ast.AST = syntax_tree

    @property
    def shadowing_name(self) -> str:
        return 'module'

    @property
    def symbols(self) -> Dict[str, ISymbol]:
        symbol = {}
        symbol.update(self.variables)
        symbol.update(self.methods)
        symbol.update(self.types)
        symbol.update(self._other_symbols)
        symbol.update(self.imports)
        return symbol

    @property
    def all_symbols(self) -> Dict[str, ISymbol]:
        symbol = self.symbols.copy()
        symbol.update(self._symbols_not_imported)
        return symbol


class BuiltinImport(Import):
    """
    A class used to differentiate built-in importings
    """
    pass



from typing import Any, Dict, List, Optional, Tuple, Union

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.expression import IExpression
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class IsInstanceMethod(IBuiltinMethod):

    def __init__(self, target_type: IType = None):
        from boa3.model.type.type import Type
        identifier = 'isinstance'

        args: Dict[str, Variable] = {
            'x': Variable(Type.any),
            'A_tuple': None
        }

        super().__init__(identifier, args, return_type=Type.bool)

        from boa3.model.type.collection.sequence.tupletype import TupleType
        from boa3.model.type.annotation.metatype import MetaType
        from boa3.model.type.annotation.uniontype import UnionType

        if not isinstance(target_type, IType):
            instances = [Type.none]
        elif isinstance(target_type, TupleType) and isinstance(target_type.item_type, UnionType):
            instances = [typ.meta_type if isinstance(typ, MetaType) else typ
                         for typ in target_type.item_type.union_types]
        else:
            instances = [target_type.meta_type if isinstance(target_type, MetaType) else target_type]

        self._instances_type: List[IType] = instances

    def set_instance_type(self, value: List[IType]):
        new_list = []
        for tpe in value:
            if isinstance(tpe, IType):
                if not any(tpe.raw_identifier == other.raw_identifier for other in new_list):
                    new_list.append(tpe)

        self._instances_type = new_list

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type
        if (len(self._instances_type) == 0
                or (len(self._instances_type) == 1 and self._instances_type[0] in (None, Type.none))
            ):
            return self._identifier

        types = list({tpe.raw_identifier for tpe in self._instances_type})
        types.sort()
        return '-{0}_of_{1}'.format(self._identifier, '_or_'.join(types))

    def args_to_be_generated(self) -> List[int]:
        args = [name for name, symbol in self.args.items() if isinstance(symbol, Variable)]
        return [list(self.args).index(key) for key in args]

    @property
    def is_supported(self) -> bool:
        from boa3.model.type.classes.classtype import ClassType
        return not any(isinstance(param, ClassType) and len(param.is_instance_opcodes()) == 0
                       for param in self._instances_type)

    def not_supported_str(self, callable_id: str) -> str:
        types = (self._instances_type[0].identifier if len(self._instances_type) == 1
                 else '({0})'.format(', '.join([arg.identifier for arg in self._instances_type])))

        return '{0}({1}, {2})'.format(callable_id, self.arg_x, types)

    @property
    def arg_x(self) -> Variable:
        return self.args['x']

    def validate_parameters(self, *params: Union[IExpression, IType]) -> bool:
        if len(params) != 2:
            return False

        return not any(not isinstance(param, (IExpression, IType)) for param in params)

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        if len(self._instances_type) == 0:
            return [
                (Opcode.ISNULL, b'')
            ]
        else:
            opcodes = []
            from boa3.model.type.type import Type
            from boa3.neo.vm.type.Integer import Integer
            types = self._instances_type.copy()

            jmps = []
            for check_instance in types[:-1]:
                opcodes.append((Opcode.DUP, b''))
                is_instance_opcodes = check_instance.is_instance_opcodes()
                opcodes.extend(is_instance_opcodes)

                jmps.append(len(opcodes))
                opcodes.append((Opcode.JMPIF, b''))

            opcodes.extend(types[-1].is_instance_opcodes())

            last_index = len(opcodes)
            if len(types) > 1:
                opcodes.extend([
                    (Opcode.JMP, Integer(4).to_byte_array(min_length=1, signed=True)),
                    (Opcode.DROP, b''),
                ])
                last_index = len(opcodes)
                opcodes.append((Opcode.PUSH1, b''))

            jmp_to = 0
            for index in reversed(jmps):
                for pos in range(index + 1, last_index):
                    last_op, last_data = opcodes[pos - 1]
                    op, data = opcodes[pos]
                    jmp_to += len(last_data) + len(op)
                jmp_to += 1

                last_index = index + 1
                opcodes[index] = opcodes[index][0], Integer(jmp_to).to_byte_array(min_length=1, signed=True)

            return opcodes

    @property
    def _args_on_stack(self) -> int:
        return 1

    @property
    def generation_order(self) -> List[int]:
        # type should not be converted
        indexes = super().generation_order
        typ_index = list(self.args).index('A_tuple')

        if typ_index in indexes:
            indexes.remove(typ_index)

        return indexes

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, list) and self.validate_parameters(*value):
            return IsInstanceMethod(value[-1])
        return super().build(value)



from typing import Any, Dict, List, Optional

from boa3 import constants
from boa3.model.callable import Callable
from boa3.model.method import Method
from boa3.model.property import Property
from boa3.model.symbol import ISymbol
from boa3.model.type.classes.classarraytype import ClassArrayType
from boa3.model.type.classes.classscope import ClassScope
from boa3.model.type.classes.classtype import ClassType
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class UserClass(ClassArrayType):
    def __init__(self, identifier: str, decorators: List[Callable] = None,
                 bases: List[ClassType] = None):
        super(ClassArrayType, self).__init__(identifier, decorators, bases)

        self._static_methods: Dict[str, Method] = {}

        self._class_variables: Dict[str, Variable] = {}
        self._class_methods: Dict[str, Method] = {}

        self._instance_variables: Dict[str, Variable] = {}
        self._instance_methods: Dict[str, Method] = {}
        self._properties: Dict[str, Property] = {}

        self.imported_symbols = {}

    @property
    def shadowing_name(self) -> str:
        return 'class'

    @property
    def class_variables(self) -> Dict[str, Variable]:
        class_vars = super().class_variables
        class_vars.update(self._class_variables)
        return class_vars

    @property
    def instance_variables(self) -> Dict[str, Variable]:
        instance_vars = super().instance_variables
        instance_vars.update(self._instance_variables)
        return instance_vars

    @property
    def properties(self) -> Dict[str, Property]:
        props = super().properties
        props.update(self._properties)
        return props

    @property
    def static_methods(self) -> Dict[str, Method]:
        static_funcs = super().static_methods
        static_funcs.update(self._static_methods)
        return static_funcs

    @property
    def class_methods(self) -> Dict[str, Method]:
        class_funcs = super().class_methods
        class_funcs.update(self._class_methods)
        return class_funcs

    @property
    def instance_methods(self) -> Dict[str, Method]:
        instance_funcs = super().instance_methods
        instance_funcs.update(self._instance_methods)
        return instance_funcs

    def include_variable(self, var_id: str, var: Variable, is_instance: bool):
        """
        Includes a variable into the list of class variables

        :param var_id: variable identifier
        :param var: variable to be included
        :param is_instance: whether is a instance variable or a class variable
        """
        if not is_instance:
            self._class_variables[var_id] = var
        else:
            self._instance_variables[var_id] = var

    def include_property(self, prop_id: str, prop: Property):
        """
        Includes a property into the list of properties

        :param prop_id: property identifier
        :param prop: property to be included
        """
        self._properties[prop_id] = prop

    def include_callable(self, method_id: str, method: Callable, scope: ClassScope = ClassScope.INSTANCE) -> bool:
        """
        Includes a method into the scope of the class

        :param method_id: method identifier
        :param method: method to be included
        :param scope: which class scope this method should be included
        """
        from boa3.model.builtin.builtin import Builtin
        if isinstance(method, Method):
            if Builtin.ClassMethodDecorator in method.decorators or scope is ClassScope.CLASS:
                methods_map = self._class_methods
            elif Builtin.StaticMethodDecorator in method.decorators or scope is ClassScope.STATIC:
                methods_map = self._static_methods
            else:
                methods_map = self._instance_methods

            if method_id not in methods_map:
                methods_map[method_id] = method
                return True

        return False

    def include_symbol(self, symbol_id: str, symbol: ISymbol, scope: ClassScope = ClassScope.INSTANCE):
        """
        Includes a method into the scope of the module

        :param symbol_id: method identifier
        :param symbol: method to be included
        :param scope: which class scope this symbol should be included
        """
        if symbol_id not in self.symbols:
            if isinstance(symbol, Variable):
                self.include_variable(symbol_id, symbol, scope == ClassScope.INSTANCE)
            elif isinstance(symbol, Property):
                self.include_property(symbol_id, symbol)
            elif isinstance(symbol, Callable):
                self.include_callable(symbol_id, symbol, scope)
            else:
                self.imported_symbols[symbol_id] = symbol

    def constructor_method(self) -> Optional[Method]:
        if constants.INIT_METHOD_ID not in self._class_methods:
            # TODO: create a generic __init__ for instantiating classes that doesn't have it declared
            from boa3.model.type.classes.classinitmethoddefault import ClassInitMethod
            self._class_methods[constants.INIT_METHOD_ID] = ClassInitMethod(self)

        return self._class_methods[constants.INIT_METHOD_ID]

    def is_type_of(self, value: Any) -> bool:
        if value is self:
            return True
        return any(base.is_type_of(value) for base in self.bases)

    @classmethod
    def _is_type_of(cls, value: Any) -> bool:
        return isinstance(value, UserClass)

    @classmethod
    def build(cls, value: Any) -> IType:
        return cls()



import ast
from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.type.primitive.ibytestringtype import IByteStringType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class StartsWithMethod(IBuiltinMethod):
    def __init__(self, self_type: IByteStringType = None):
        from boa3.model.type.type import Type

        if not isinstance(self_type, IByteStringType):
            from boa3.model.type.primitive.bytestringtype import ByteStringType
            self_type = ByteStringType.build()

        identifier = 'startswith'
        args: Dict[str, Variable] = {
            'self': Variable(self_type),
            'value': Variable(self_type),
            'start': Variable(Type.int),
            'end': Variable(Type.int),
        }

        start_default = ast.parse("{0}".format(0)
                                  ).body[0].value
        end_default = ast.parse("-1").body[0].value.operand
        end_default.n = -1

        super().__init__(identifier, args, defaults=[start_default, end_default], return_type=Type.bool)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count
        from boa3.neo.vm.type.StackItem import StackItemType

        jmp_place_holder = (Opcode.JMP, b'\x01')
        # string, end, start, substring

        verify_negative_index = [           # verify if index is negative value
            (Opcode.DUP, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder                # if not index < 0, jump to verify_big_end
        ]

        fix_negative_end = [                # fix negative value to a positive equivalent
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.ADD, b''),
            (Opcode.INC, b''),              # end = end + len(string) + 1
            (Opcode.DUP, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder                # if not end < 0, jump to verify_and_fix_start
        ]

        fix_still_negative_index = [        # if end still is less than 0, then it should be 0
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),            # end = 0
        ]

        jmp_fix_negative_index = Opcode.get_jump_and_data(Opcode.JMPGT, get_bytes_count(fix_negative_end +
                                                                                        fix_still_negative_index), True)
        verify_negative_index[-1] = jmp_fix_negative_index

        verify_big_end = [                  # verify if end is greater or equals to len(string)
            (Opcode.DUP, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            jmp_place_holder                # if not end >= len(string), jump to verify_and_fix_start
        ]

        fix_big_end = [                     # fix end value to a positive equivalent
            (Opcode.DROP, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),             # end = len(string)
        ]

        jmp_other_verifies = Opcode.get_jump_and_data(Opcode.JMPGT, get_bytes_count(fix_still_negative_index +
                                                                                    verify_big_end +
                                                                                    fix_big_end), True)
        fix_negative_end[-1] = jmp_other_verifies

        jmp_fix_big_index = Opcode.get_jump_and_data(Opcode.JMPLE, get_bytes_count(fix_big_end), True)
        verify_big_end[-1] = jmp_fix_big_index

        verify_and_fix_end = [              # verify and fix end index
            (Opcode.REVERSE3, b''),         # change positions on stack
        ]
        verify_and_fix_end.extend(verify_negative_index)
        verify_and_fix_end.extend(fix_negative_end)
        verify_and_fix_end.extend(fix_still_negative_index)
        verify_and_fix_end.extend(verify_big_end)
        verify_and_fix_end.extend(fix_big_end)

        verify_and_fix_start = [            # verify and fix end index
            (Opcode.SWAP, b''),             # change positions on stack
        ]

        # fix_negative_start is the same as fix_negative_end, but jump is different
        fix_negative_start = fix_negative_end.copy()
        verify_big_start = [                # verify if start is greater or equals to len(string)
            (Opcode.DUP, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            jmp_place_holder                # if start >= len(string), return False
        ]

        jmp_other_verifies = Opcode.get_jump_and_data(Opcode.JMPGT, get_bytes_count(fix_still_negative_index +
                                                                                    verify_big_start), True)
        fix_negative_end[-1] = jmp_other_verifies

        verify_size = [                     # verify if len(string[start:end]) > len(substring)
            (Opcode.REVERSE3, b''),
            (Opcode.REVERSE4, b''),
            (Opcode.REVERSE3, b''),
            (Opcode.OVER, b''),
            (Opcode.SUB, b''),
            (Opcode.SUBSTR, b''),

            (Opcode.PUSH0, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.DUP, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            jmp_place_holder                # if len(string[start:end]) > len(substring), return False
        ]

        compare_starts = [                  # string.startswith(substring)
            (Opcode.SUBSTR, b''),
            (Opcode.CONVERT, StackItemType.ByteString),
            (Opcode.NUMEQUAL, b''),         # return string[start:end][0:len(substring)] == substring
            jmp_place_holder                # jumps other opcodes
        ]

        jmp_compare_starts = Opcode.get_jump_and_data(Opcode.JMPGT, get_bytes_count(compare_starts), True)
        verify_size[-1] = jmp_compare_starts

        jmp_to_false = Opcode.get_jump_and_data(Opcode.JMPGE, get_bytes_count(verify_size +
                                                                              compare_starts), True)
        verify_big_start[-1] = jmp_to_false

        verify_and_fix_start.extend(verify_negative_index)
        verify_and_fix_start.extend(fix_negative_start)
        verify_and_fix_start.extend(fix_still_negative_index)
        verify_and_fix_start.extend(verify_big_start)

        return_false = [                    # remove all values from stack and put False
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),            # return False
        ]

        jmp_bigger_substr = Opcode.get_jump_and_data(Opcode.JMP, get_bytes_count(return_false), True)
        compare_starts[-1] = jmp_bigger_substr

        return (
            verify_and_fix_end +
            verify_and_fix_start +
            verify_size +
            compare_starts +
            return_false
        )

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, IByteStringType):
            return StartsWithMethod(value)
        return super().build(value)



import ast
from typing import Dict

from boa3.model.builtin.interop.nativecontract import StdLibMethod
from boa3.model.variable import Variable


class AtoiMethod(StdLibMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'atoi'
        syscall = 'atoi'
        args: Dict[str, Variable] = {
            'value': Variable(Type.str),
            'base': Variable(Type.int)
        }

        args_default = ast.parse("{0}".format(10)
                                 ).body[0].value

        super().__init__(identifier, syscall, args, defaults=[args_default], return_type=Type.int)



__all__ = ['AbsMethod',
           'BoolMethod',
           'IBuiltinMethod',
           'ByteArrayMethod',
           'ByteArrayEncodingMethod',
           'CreateEventMethod',
           'EventType',
           'ExceptionMethod',
           'ExitMethod',
           'IntByteStringMethod',
           'IntIntMethod',
           'IsInstanceMethod',
           'LenMethod',
           'ListBytesStringMethod',
           'ListGenericMethod',
           'ListMappingMethod',
           'ListSequenceMethod',
           'MaxIntMethod',
           'MaxByteStringMethod',
           'MinByteStringMethod',
           'MinIntMethod',
           'PrintMethod',
           'RangeMethod',
           'ReversedMethod',
           'ScriptHashMethod',
           'StrSplitMethod',
           'SumMethod',
           'SuperMethod',
           ]

from boa3.model.builtin.method.absmethod import AbsMethod
from boa3.model.builtin.method.boolmethod import BoolMethod
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.builtin.method.bytearraymethod import ByteArrayMethod
from boa3.model.builtin.method.bytearrayencodingmethod import ByteArrayEncodingMethod
from boa3.model.builtin.method.createeventmethod import CreateEventMethod, EventType
from boa3.model.builtin.method.exceptionmethod import ExceptionMethod
from boa3.model.builtin.method.exitmethod import ExitMethod
from boa3.model.builtin.method.intbytestringmethod import IntByteStringMethod
from boa3.model.builtin.method.intintmethod import IntIntMethod
from boa3.model.builtin.method.isinstancemethod import IsInstanceMethod
from boa3.model.builtin.method.lenmethod import LenMethod
from boa3.model.builtin.method.listbytesstringmethod import ListBytesStringMethod
from boa3.model.builtin.method.listgenericmethod import ListGenericMethod
from boa3.model.builtin.method.listmappingmethod import ListMappingMethod
from boa3.model.builtin.method.listsequencemethod import ListSequenceMethod
from boa3.model.builtin.method.maxbytestringmethod import MaxByteStringMethod
from boa3.model.builtin.method.maxintmethod import MaxIntMethod
from boa3.model.builtin.method.minbytestringmethod import MinByteStringMethod
from boa3.model.builtin.method.minintmethod import MinIntMethod
from boa3.model.builtin.method.printmethod import PrintMethod
from boa3.model.builtin.method.rangemethod import RangeMethod
from boa3.model.builtin.method.reversedmethod import ReversedMethod
from boa3.model.builtin.method.strsplitmethod import StrSplitMethod
from boa3.model.builtin.method.summethod import SumMethod
from boa3.model.builtin.method.supermethod import SuperMethod
from boa3.model.builtin.method.toscripthashmethod import ScriptHashMethod



from __future__ import annotations

from typing import Any, Dict, Optional

from boa3.model.method import Method
from boa3.model.property import Property
from boa3.model.type.classes.classarraytype import ClassArrayType
from boa3.model.variable import Variable


class CryptoLibClass(ClassArrayType):
    """
    A class used to represent CryptoLib native contract
    """

    def __init__(self):
        super().__init__('CryptoLib')

        self._variables: Dict[str, Variable] = {}
        self._class_methods: Dict[str, Method] = {}
        self._constructor: Optional[Method] = None

    @property
    def instance_variables(self) -> Dict[str, Variable]:
        return self._variables.copy()

    @property
    def class_variables(self) -> Dict[str, Variable]:
        return {}

    @property
    def properties(self) -> Dict[str, Property]:
        return {}

    @property
    def static_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def class_methods(self) -> Dict[str, Method]:
        # avoid recursive import
        from boa3.model.builtin.interop.crypto import (Murmur32Method, Sha256Method, Ripemd160Method,
                                                       VerifyWithECDsaMethod)

        if len(self._class_methods) == 0:
            self._class_methods = {
                'murmur32': Murmur32Method(),
                'sha256': Sha256Method(),
                'ripemd160': Ripemd160Method(),
                'verify_with_ecdsa': VerifyWithECDsaMethod(),
            }
        return self._class_methods

    @property
    def instance_methods(self) -> Dict[str, Method]:
        return {}

    def constructor_method(self) -> Optional[Method]:
        return self._constructor

    @classmethod
    def build(cls, value: Any = None) -> CryptoLibClass:
        if value is None or cls._is_type_of(value):
            return _CryptoLib

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, CryptoLibClass)


_CryptoLib = CryptoLibClass()



import ast
from typing import Dict, List, Tuple

from boa3.model.builtin.method.intmethod import IntMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class IntIntMethod(IntMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        args: Dict[str, Variable] = {
            'value': Variable(Type.int),
        }

        value_default = ast.parse("{0}".format(Type.int.default_value)
                                  ).body[0].value

        super().__init__(args, [value_default])

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        # returns the same int
        return [(Opcode.RET, b'')]



import ast
from typing import Any, Callable, Dict, List, Optional, Type

from boa3.analyser.astanalyser import IAstAnalyser
from boa3.exception import CompilerError
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.builtin.method.isinstancemethod import IsInstanceMethod
from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType


class BuiltinFunctionCallAnalyser(IAstAnalyser):
    def __init__(self, origin: IAstAnalyser, call: ast.Call, method_id: str, builtin_method: IBuiltinMethod, log: bool):
        self._method: IBuiltinMethod = builtin_method
        self.method_id: str = method_id
        super().__init__(call, log=log)

        self._origin: IAstAnalyser = origin

        # all methods validators must be (IBuiltinMethod, List[IType]) -> None
        self._methods_validators: Dict[Type[IBuiltinMethod],
                                       Callable[[IBuiltinMethod, List[IType]], None]] = {
            IsInstanceMethod: self._validate_IsInstanceMethod
        }

    @property
    def method(self) -> IBuiltinMethod:
        return self._method

    @property
    def call(self) -> ast.Call:
        return self._tree

    def get_symbol(self, symbol_id: str,
                   is_internal: bool = False,
                   check_raw_id: bool = False) -> Optional[ISymbol]:
        return self._origin.get_symbol(symbol_id, is_internal)

    def get_symbol_from_node(self, node: ast.AST) -> Optional[ISymbol]:
        if isinstance(node, ast.Name):
            return self.get_symbol(node.id)

        if isinstance(node, ast.Attribute):
            value = self.get_symbol_from_node(node.value)
            if hasattr(value, 'symbols') and node.attr in value.symbols:
                return value.symbols[node.attr]

        return None

    def get_type(self, value: Any, use_metadata: bool = False) -> IType:
        return self._origin.get_type(value, use_metadata)

    def validate(self) -> bool:
        """
        Validates the method arguments.

        :return: whether the method have specific validation
        """
        if type(self.method) in self._methods_validators:
            if self.method_id == self.method.raw_identifier:
                # if the identifiers are diffrente, this method was validated already
                validator = self._methods_validators[type(self.method)]
                args = [self.get_type(param) for param in self.call.args]
                validator(self.method, args)
            return True
        return False

    def _validate_IsInstanceMethod(self, method: IsInstanceMethod, args_types: List[IType]):
        """
        Validates the arguments for `isinstance` method

        :param method: instance of the builtin method
        :param args_types: types of the arguments
        """
        last_arg = self.call.args[-1]
        if isinstance(last_arg, ast.Tuple) and all(isinstance(tpe, (ast.Attribute, ast.Name))
                                                   for tpe in last_arg.elts):
            if len(last_arg.elts) == 1:
                # if the types tuple has only one type, remove it from inside the tuple
                last_arg = self.call.args[-1] = last_arg.elts[-1]
                args_types[-1] = args_types[-1].value_type
            elif len(last_arg.elts) > 1:
                # if there are more than one type, updates information in the instance of the method
                types: List[IType] = [self.get_symbol_from_node(name) for name in last_arg.elts]
                method.set_instance_type(types)
                self.call.args[-1] = last_arg.elts[-1]
                return

        from boa3.model.type.annotation.metatype import MetaType
        from boa3.model.type.type import Type
        is_ast_valid = (isinstance(last_arg, ast.Name)
                        or (isinstance(last_arg, ast.NameConstant) and args_types[-1] is Type.none))

        is_id_valid = (hasattr(last_arg, 'id')
                       and last_arg.id != args_types[-1].identifier
                       and last_arg.id != args_types[-1].raw_identifier
                       and isinstance(self.get_type(last_arg, use_metadata=True), MetaType))

        if not is_ast_valid or is_id_valid:
            # if the value is not the identifier of a type
            self._log_error(
                CompilerError.MismatchedTypes(
                    last_arg.lineno, last_arg.col_offset,
                    expected_type_id=type.__name__,
                    actual_type_id=args_types[-1].identifier
                ))



from boa3.model.builtin.decorator.builtindecorator import IBuiltinDecorator


class StaticMethodDecorator(IBuiltinDecorator):
    def __init__(self):
        identifier = 'staticmethod'
        super().__init__(identifier)



import abc
from typing import Any, List

from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.itype import IType
from boa3.model.type.primitive.primitivetype import PrimitiveType
from boa3.neo.vm.type.AbiType import AbiType
from boa3.neo.vm.type.StackItem import StackItemType


class IByteStringType(SequenceType, PrimitiveType, abc.ABC):
    """
    A class used to represent Neo ByteString type
    """

    def __init__(self, identifier: str, values_type: List[IType]):
        super().__init__(identifier, values_type)

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def abi_type(self) -> AbiType:
        return AbiType.ByteArray

    @property
    def stack_item(self) -> StackItemType:
        return StackItemType.ByteString

    def is_type_of(self, value: Any) -> bool:
        return self._is_type_of(value)

    def _init_class_symbols(self):
        super()._init_class_symbols()

        from boa3.model.builtin.builtin import Builtin

        instance_methods = [Builtin.BytesStringIsDigit,
                            Builtin.BytesStringJoin,
                            Builtin.BytesStringLower,
                            Builtin.BytesStringUpper,
                            Builtin.BytesStringStartswith,
                            Builtin.BytesStringStrip,
                            ]

        for instance_method in instance_methods:
            self._instance_methods[instance_method.raw_identifier] = instance_method.build(self)

    def is_valid_key(self, key_type: IType) -> bool:
        return key_type == self.valid_key

    @property
    def valid_key(self) -> IType:
        from boa3.model.type.type import Type
        return Type.int

    @property
    def can_reassign_values(self) -> bool:
        return False



from __future__ import annotations

from typing import Dict, List, Tuple, Union

from boa3.compiler.codegenerator.vmcodemapping import VMCodeMapping
from boa3.model.type.itype import IType
from boa3.neo.vm.VMCode import VMCode


class StackMemento:
    """
    This class is responsible for managing the simulation of the blockchain stack during the code generation
    """

    def __init__(self):
        self._stacks: List[Tuple[VMCode, NeoStack]] = []
        self._current_stack: NeoStack = NeoStack()

    @property
    def stack_map(self) -> Dict[int, NeoStack]:
        vm_code_mapping = VMCodeMapping.instance()
        return {vm_code_mapping.get_start_address(vmcode): stack for vmcode, stack in self._stacks}

    def get_state(self, code_address: int) -> NeoStack:
        stacks = self.stack_map
        index = None
        for address, stack in sorted(stacks.items()):
            if address >= code_address:
                break
            index = address

        if index is None or index not in stacks:
            return NeoStack()
        else:
            return stacks[index]

    @property
    def current_stack(self) -> NeoStack:
        return self._current_stack

    def append(self, value: IType, code: VMCode):
        states = self.stack_map
        index = VMCodeMapping.instance().get_start_address(code)
        if index in states:
            states[index].append(value)

        else:
            if self._current_stack is not None:
                stack = self._current_stack.copy()
            else:
                stack = NeoStack()
            stack.append(value)

            self._stacks.append((code, stack))
            self._current_stack = stack


class NeoStack:
    def __init__(self):
        self._stack: List[IType] = []

    def append(self, value: IType):
        return self._stack.append(value)

    def clear(self):
        return self._stack.clear()

    def copy(self) -> NeoStack:
        new_stack = NeoStack()
        new_stack._stack = self._stack.copy()
        return new_stack

    def pop(self, index: int) -> IType:
        return self._stack.pop(index)

    def __len__(self) -> int:
        return len(self._stack)

    def __getitem__(self, index_or_slice: Union[int, slice]):
        return self._stack[index_or_slice]

    def reverse(self, start: int = 0, end: int = None):
        if end is None:
            end = len(self._stack)

        reverse = list(reversed(self._stack[start:end]))
        self._stack[start:end] = reverse



from __future__ import annotations

from typing import Any, Dict, Optional

from boa3.model.method import Method
from boa3.model.property import Property
from boa3.model.type.classes.classarraytype import ClassArrayType
from boa3.model.variable import Variable


class StdLibClass(ClassArrayType):
    """
    A class used to represent StdLib native contract
    """

    def __init__(self):
        super().__init__('StdLib')

        self._variables: Dict[str, Variable] = {}
        self._class_methods: Dict[str, Method] = {}
        self._constructor: Optional[Method] = None

    @property
    def instance_variables(self) -> Dict[str, Variable]:
        return self._variables.copy()

    @property
    def class_variables(self) -> Dict[str, Variable]:
        return {}

    @property
    def properties(self) -> Dict[str, Property]:
        return {}

    @property
    def static_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def class_methods(self) -> Dict[str, Method]:
        # avoid recursive import
        from boa3.model.builtin.interop.stdlib import (SerializeMethod, DeserializeMethod,
                                                       Base64DecodeMethod, Base64EncodeMethod,
                                                       Base58DecodeMethod, Base58EncodeMethod,
                                                       Base58CheckDecodeMethod, Base58CheckEncodeMethod,
                                                       ItoaMethod, AtoiMethod,
                                                       MemoryCompareMethod, MemorySearchMethod)
        from boa3.model.builtin.interop.json import (JsonDeserializeMethod, JsonSerializeMethod)

        if len(self._class_methods) == 0:
            self._class_methods = {
                'serialize': SerializeMethod(),
                'deserialize': DeserializeMethod(),
                'json_serialize': JsonSerializeMethod(),
                'json_deserialize': JsonDeserializeMethod(),
                'base64_decode': Base64DecodeMethod(),
                'base64_encode': Base64EncodeMethod(),
                'base58_decode': Base58DecodeMethod(),
                'base58_encode': Base58EncodeMethod(),
                'base58_check_decode': Base58CheckDecodeMethod(),
                'base58_check_encode': Base58CheckEncodeMethod(),
                'itoa': ItoaMethod(),
                'atoi': AtoiMethod(),
                'memory_compare': MemoryCompareMethod(),
                'memory_search': MemorySearchMethod(),
            }
        return self._class_methods

    @property
    def instance_methods(self) -> Dict[str, Method]:
        return {}

    def constructor_method(self) -> Optional[Method]:
        return self._constructor

    @classmethod
    def build(cls, value: Any = None) -> StdLibClass:
        if value is None or cls._is_type_of(value):
            return _StdLib

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, StdLibClass)


_StdLib = StdLibClass()



from __future__ import annotations

from typing import Sequence, Union


class UInt160(bytes):
    """
    Represents a 160-bit unsigned integer.
    """

    def __init__(self, arg: Union[bytes, int] = 0):
        super().__init__()
        pass


class UInt256(bytes):
    """
    Represents a 256-bit unsigned integer.
    """

    def __init__(self, arg: Union[bytes, int] = 0):
        super().__init__()
        pass


class ECPoint(bytes):
    """
    Represents a coordinate pair for elliptic curve cryptography (ECC) structures.
    """

    def __init__(self, arg: bytes):
        super().__init__()
        pass

    def to_script_hash(self) -> bytes:
        """
        Converts a data to a script hash.

        :return: the script hash of the data
        :rtype: bytes
        """
        pass


class ByteString:
    """
    An type annotation for values that can be str or bytes. Same as Union[str, bytes]
    """

    def to_bytes(self) -> bytes:
        """
        Converts an ByteString value to an array of bytes
        """
        pass

    def to_str(self) -> str:
        """
        Converts an ByteString value to a string.
        """
        pass

    def to_int(self) -> int:
        """
        Return the integer represented by this ByteString.
        """
        pass

    def to_bool(self) -> bool:
        """
        Return the boolean represented by this ByteString.
        """
        pass

    def isdigit(self) -> bool:
        """
        Return True if the ByteString is a digit string, False otherwise.

        A ByteString is a digit string if all characters in the ByteString are digits and there
        is at least one character in the ByteString.
        """
        pass

    def join(self, __iterable: Sequence[ByteString]) -> ByteString:
        """
        S.join(__iterable: Sequence[ByteString]) -> ByteString

        Concatenate any number of ByteStrings.

        The ByteString whose method is called is inserted in between each given ByteString.
        The result is returned as a new ByteString.
        """
        pass

    def lower(self) -> ByteString:
        """
        S.lower() -> ByteString

        Return a copy of the ByteString converted to lowercase.
        """
        pass

    def startswith(self, prefix: ByteString, start: int = 0, end: int = -1) -> bool:
        """
        S.startswith(prefix: ByteString, start: int, end: int) -> bool

        Return True if S starts with the specified prefix, False otherwise.
        With optional start, test S beginning at that position.
        With optional end, stop comparing S at that position.
        prefix can also be a tuple of strings to try.
        """
        pass

    def strip(self, __chars: ByteString) -> ByteString:
        """
        S.strip(__chars: ByteString) -> ByteString

        Return a copy of the ByteString with leading and trailing whitespace remove.

        If chars is given and not None, remove characters in chars instead.
        """
        pass

    def upper(self) -> ByteString:
        """
        S.upper() -> ByteString

        Return a copy of the ByteString converted to uppercase.
        """
        pass

    def __add__(self, *args, **kwargs):  # real signature unknown
        """
        Return self+value.
        """
        pass

    def __mul__(self, *args, **kwargs):  # real signature unknown
        """
        Return self*value.
        """
        pass

    def __rmul__(self, *args, **kwargs):  # real signature unknown
        """
        Return value*self.
        """
        pass



import ast
from typing import Any, Dict, List, Optional, Sized

from boa3.model.builtin.decorator.builtindecorator import IBuiltinDecorator
from boa3.model.type.type import Type
from boa3.model.variable import Variable


class PublicDecorator(IBuiltinDecorator):
    def __init__(self):
        identifier = 'public'
        args: Dict[str, Variable] = {'name': Variable(Type.str),
                                     'safe': Variable(Type.bool),
                                     }

        name_default = ast.parse("'{0}'".format(Type.str.default_value)).body[0].value
        safe_default = ast.parse("{0}".format(Type.bool.default_value)).body[0].value

        defaults = [name_default, safe_default]
        super().__init__(identifier, args, defaults)

        self.name: Optional[str] = None
        self.safe = False

    def build(self, *args) -> IBuiltinDecorator:
        if isinstance(args, Sized) and len(args) > 0 and isinstance(args[0], ast.AST):
            decorator = PublicDecorator()
            origin = args[0]
            decorator._origin_node = origin

            if len(args) > 1:
                values = self.validate_values(*args[:2])

                name_arg, safe_arg = values
                if isinstance(name_arg, str) and len(name_arg) > 0:
                    decorator.name = name_arg
                if isinstance(safe_arg, bool):
                    decorator.safe = safe_arg

            return decorator

        return self

    def validate_values(self, *params: Any) -> List[Any]:
        values = []
        if len(params) != 2:
            return values

        origin, visitor = params
        from boa3.analyser.astanalyser import IAstAnalyser
        if not isinstance(origin, ast.Call) or not isinstance(visitor, IAstAnalyser):
            return [self.name, self.safe]

        # read the called arguments
        args_names = list(self.args.keys())
        args_len = min(len(origin.args), len(args_names))
        for x in range(args_len):
            values.append(visitor.visit(origin.args[x]))

        if len(values) <= 1:
            if len(values) == 0:
                values.append('')
            values.append(self.safe)

        # read the called keyword arguments
        for kwarg in origin.keywords:
            if kwarg.arg in args_names:
                x = args_names.index(kwarg.arg)
                value = visitor.visit(kwarg.value)
                values[x] = value

        from boa3.exception import CompilerError
        name, safe = values
        if not isinstance(name, str):
            visitor._log_error(
                CompilerError.MismatchedTypes(
                    origin.lineno, origin.col_offset,
                    expected_type_id=Type.str.identifier,
                    actual_type_id=type(name).__name__
                ))

        if not isinstance(safe, bool):
            visitor._log_error(
                CompilerError.MismatchedTypes(
                    origin.lineno, origin.col_offset,
                    expected_type_id=Type.bool.identifier,
                    actual_type_id=type(safe).__name__
                ))

        return values



import abc
from typing import List, Optional, Union

from boa3.model.event import Event
from boa3.model.method import Method


class INeoStandard(abc.ABC):
    def __init__(self, methods: List[Method], events: List[Event], optionals: Optional[List[Method]] = None):
        if optionals is None:
            optionals = []
        self.methods: List[Method] = methods.copy()
        self.events: List[Event] = events.copy()
        self.optionals: List[Method] = optionals.copy()

    def match_definition(self, standard: Union[Method, Event], symbol: Union[Method, Event]) -> bool:
        if not isinstance(standard, (Method, Event)) or not isinstance(symbol, (Method, Event)):
            return False

        standard_symbols = self.methods + self.optionals if isinstance(symbol, Method) else self.events
        if standard not in standard_symbols:
            return False

        return self._have_same_signature(standard, symbol)

    def _have_same_signature(self, symbol: Union[Method, Event], other: Union[Method, Event]) -> bool:
        if (isinstance(symbol, Method) and not isinstance(other, Method)
                or isinstance(symbol, Event) and not isinstance(other, Event)):
            return False

        if symbol.return_type != other.return_type and not (
                # verifies if both return types are equal if they are a neo3-boa type
                symbol.return_type.is_type_of(other.return_type) and other.return_type.is_type_of(symbol.return_type)):
            return False

        if symbol.is_public != other.is_public:
            return False

        if symbol.is_safe != other.is_safe:
            return False

        if len(symbol.args) != len(other.args):
            return False

        method_args = list(symbol.args.values())
        other_args = list(other.args.values())
        for index in range(len(method_args)):
            if method_args[index].type != other_args[index].type:
                return False

        return True
